"""
This module is designed to make STCAL tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


STCAL copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/stcal/blob/main/src/stcal/alignment/util.py
"""

from __future__ import annotations

import functools
import logging
import re
from typing import TYPE_CHECKING
import warnings

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence

    import astropy

import gwcs
import numpy as np
from astropy import wcs as fitswcs
from astropy.coordinates import SkyCoord
from astropy.modeling import models as astmodels
from astropy.utils.misc import isiterable
from gwcs.wcstools import wcs_from_fiducial

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

__all__ = [
    "compute_scale",
    "compute_fiducial",
    "calc_rotation_matrix",
    "compute_s_region_imaging",
    "compute_s_region_keyword",
    "wcs_from_footprints",
    "wcs_from_sregions",
    "wcs_bbox_from_shape",
    "reproject",
]


def _calculate_fiducial_from_spatial_footprint(
    spatial_footprint: np.ndarray,
) -> tuple:
    """
    Calculates the fiducial coordinates from a given spatial footprint.

    Parameters
    ----------
    spatial_footprint : np.ndarray
        An Nx2 array containing the world coordinates of the WCS footprint's
        bounding box, where N is the number of bounding box positions.

    Returns
    -------
    lon_fiducial, lat_fiducial : np.ndarray, np.ndarray
        The world coordinates of the fiducial point in the output coordinate frame.
    """
    lon, lat = spatial_footprint.T
    lon, lat = np.deg2rad(lon), np.deg2rad(lat)
    x = np.cos(lat) * np.cos(lon)
    y = np.cos(lat) * np.sin(lon)
    z = np.sin(lat)

    x_mid = (np.max(x) + np.min(x)) / 2.0
    y_mid = (np.max(y) + np.min(y)) / 2.0
    z_mid = (np.max(z) + np.min(z)) / 2.0
    lon_fiducial = np.rad2deg(np.arctan2(y_mid, x_mid)) % 360.0
    lat_fiducial = np.rad2deg(np.arctan2(z_mid, np.sqrt(x_mid**2 + y_mid**2)))
    return lon_fiducial, lat_fiducial


def _generate_tranform(
    wcs: gwcs.wcs.WCS,
    wcsinfo: dict,
    ref_fiducial: np.ndarray,
    pscale_ratio: float | None = None,
    pscale: float | None = None,
    rotation: float | None = None,
    transform: astmodels.Model | None = None,
) -> astmodels.Model:
    """
    Creates a transform from pixel to world coordinates based on a
    reference datamodel's WCS.

    Parameters
    ----------
    wcs : ~gwcs.wcs.WCS
        The WCS object.

    wcsinfo : dict
        A dictionary containing the WCS FITS keywords and corresponding values.

    pscale_ratio : int, None
        Ratio of input to output pixel scale. This parameter is only used when
        ``pscale=None`` and, in that case, it is passed on to ``compute_scale``.

    pscale : float, None
        The plate scale. If `None`, the plate scale is calculated from the reference
        datamodel.

    rotation : float, None
        Position angle of output image's Y-axis relative to North.
        A value of 0.0 would orient the final output image to be North up.
        The default of `None` specifies that the images will not be rotated,
        but will instead be resampled in the default orientation for the camera
        with the x and y axes of the resampled image corresponding
        approximately to the detector axes. Ignored when ``transform`` is
        provided. If `None`, the rotation angle is extracted from the
        reference model's ``meta.wcsinfo.roll_ref``.

    ref_fiducial : np.array
        A two-elements array containing the world coordinates of the fiducial point.

    transform : ~astropy.modeling.Model
        A transform between frames.

    Returns
    -------
    transform : ~astropy.modeling.Model
        An :py:mod:`~astropy` model containing the transform between frames.
    """
    if transform is None:
        sky_axes = wcs._get_axes_indices().tolist()  # noqa: SLF001
        v3yangle = np.deg2rad(wcsinfo["v3yangle"])
        vparity = wcsinfo["vparity"]
        if rotation is None:
            roll_ref = np.deg2rad(wcsinfo["roll_ref"])
        else:
            roll_ref = np.deg2rad(rotation) + (vparity * v3yangle)

        # reshape the rotation matrix returned from calc_rotation_matrix
        # into the correct shape for constructing the transformation
        pc = np.reshape(calc_rotation_matrix(roll_ref, v3yangle, vparity=vparity), (2, 2))

        rotation = astmodels.AffineTransformation2D(pc, name="pc_rotation_matrix")
        transform = [rotation]
        if sky_axes:
            if not pscale:
                pscale = compute_scale(wcs, ref_fiducial, pscale_ratio=pscale_ratio)
            transform.append(astmodels.Scale(pscale, name="cdelt1") & astmodels.Scale(pscale, name="cdelt2"))

        if transform:
            transform = functools.reduce(lambda x, y: x | y, transform)

    return transform


def _get_axis_min_and_bounding_box(footprints: list[np.ndarray],
                                   ref_wcs: gwcs.wcs.WCS) -> tuple:
    """
    Calculates axis minimum values and bounding box.

    Parameters
    ----------
    footprints : list
        A list of numpy arrays each of shape (N, 2) containing the
        (RA, Dec) vertices demarcating the footprint of the input WCSs.

    ref_wcs : ~gwcs.wcs.WCS
        The reference WCS object.

    Returns
    -------
    tuple
        A tuple containing two elements:
            1 - a :py:class:`np.ndarray` with the minimum value in each axis;
            2 - a tuple containing the bounding box region in the format
            ((x0_lower, x0_upper), (x1_lower, x1_upper)).
    """
    domain_bounds = np.hstack([ref_wcs.backward_transform(*f.T) for f in footprints])
    axis_min_values = np.min(domain_bounds, axis=1)
    domain_bounds = (domain_bounds.T - axis_min_values).T

    output_bounding_box = []
    for axis in ref_wcs.output_frame.axes_order:
        axis_min, axis_max = (
            domain_bounds[axis].min(),
            domain_bounds[axis].max(),
        )
        # populate output_bounding_box
        output_bounding_box.append((axis_min, axis_max))

    return (axis_min_values, output_bounding_box)


def _calculate_fiducial(footprints: list[np.ndarray],
                        crval: Sequence | None = None) -> tuple:
    """
    Calculates the coordinates of the fiducial point and, if necessary, updates it with
    the values in CRVAL (the update is applied to spatial axes only).

    Parameters
    ----------
    footprints : list
        A list of numpy arrays each of shape (N, 2) containing the
        (RA, Dec) vertices demarcating the footprint of the input WCSs.

    crval : list, optional
        A reference world coordinate associated with the reference pixel. If not `None`,
        then the fiducial coordinates of the spatial axes will be updated with the
        values from ``crval``.

    Returns
    -------
    fiducial : tuple
        A tuple containing the world coordinate of the fiducial point.
    """
    if crval is not None:
        return tuple(crval)
    return _compute_fiducial_from_footprints(footprints)


def _calculate_offsets(fiducial: tuple,
                       wcs: gwcs.wcs.WCS | None,
                       axis_min_values: np.ndarray | None,
                       crpix: Sequence | None) -> astmodels.Model:
    """
    Calculates the offsets to the transform.

    Parameters
    ----------
    fiducial : tuple
        A tuple containing the world coordinates of the fiducial point.

    wcs : ~gwcs.wcs.WCS
        A WCS object. It will be used to determine the

    axis_min_values : np.ndarray
        A two-elements array containing the minimum pixel value for each axis.

    crpix : list or tuple
        Pixel coordinates of the reference pixel.

    Returns
    -------
    ~astropy.modeling.Model
        A model with the offsets to be added to the WCS's transform.

    Notes
    -----
    If ``crpix=None``, then ``fiducial``, ``wcs``, and ``axis_min_values`` must be
    provided, in which case, the offsets will be calculated using the WCS object to
    find the pixel coordinates of the fiducial point and then correct it by the minimum
    pixel value for each axis.
    """
    if crpix is None and fiducial is not None and wcs is not None and axis_min_values is not None:
        offset1, offset2 = wcs.backward_transform(*fiducial)
        offset1 -= axis_min_values[0]
        offset2 -= axis_min_values[1]
    elif crpix is None:
        msg = "If crpix is not provided, fiducial, wcs, and axis_min_values must be provided."
        raise ValueError(msg)
    else:
        offset1, offset2 = crpix

    return astmodels.Shift(-offset1, name="crpix1") & astmodels.Shift(-offset2, name="crpix2")


def _calculate_new_wcs(wcs: gwcs.wcs.WCS,
                       shape: Sequence | None,
                       footprints: list[np.ndarray],
                       fiducial: tuple,
                       crpix: Sequence | None = None,
                       transform: astmodels.Model | None = None,
                       ) -> gwcs.wcs.WCS:
    """
    Calculates a new WCS object based on the combined footprints
    and reference WCS provided.

    Parameters
    ----------
    wcs : ~gwcs.wcs.WCS
        The reference WCS object.

    shape : list
        The shape of the new WCS's pixel grid. If `None`, then the output bounding box
        will be used to determine it.

    footprints : list
        A list of numpy arrays each of shape (N, 2) containing the
        (RA, Dec) vertices demarcating the footprint of the input WCSs.

    fiducial : tuple
        A tuple containing the location on the sky in some standard
        coordinate system.

    crpix : tuple, optional
        The coordinates of the reference pixel.

    transform : ~astropy.modeling.Model
        An optional transform to be prepended to the transform constructed by the
        fiducial point. The number of outputs of this transform must equal the number
        of axes in the coordinate frame.

    Returns
    -------
    wcs_new : ~gwcs.wcs.WCS
        The new WCS object that corresponds to the combined WCS objects in `wcs_list`.
    """
    wcs_new = wcs_from_fiducial(
        fiducial,
        coordinate_frame=wcs.output_frame,
        projection=astmodels.Pix2Sky_TAN(),
        transform=transform,
        input_frame=wcs.input_frame,
    )
    axis_min_values, output_bounding_box = _get_axis_min_and_bounding_box(footprints, wcs_new)
    offsets = _calculate_offsets(
        fiducial=fiducial,
        wcs=wcs_new,
        axis_min_values=axis_min_values,
        crpix=crpix,
    )

    wcs_new.insert_transform("detector", offsets, after=True)
    wcs_new.bounding_box = output_bounding_box

    if shape is None:
        shape = [int(axs[1] - axs[0] + 0.5) for axs in output_bounding_box[::-1]]

    wcs_new.pixel_shape = shape[::-1]
    wcs_new.array_shape = shape
    return wcs_new


def _validate_wcs_list(wcs_list: list[gwcs.wcs.WCS]) -> bool:
    """
    Validates wcs_list.

    Parameters
    ----------
    wcs_list : list
        A list of WCS objects.

    Returns
    -------
    bool or Exception
        If wcs_list is valid, returns True. Otherwise, it will raise an error.

    Raises
    ------
    ValueError
        Raised whenever wcs_list is not an iterable.
    TypeError
        Raised whenever wcs_list is empty or any of its content is not an
        instance of WCS.
    """
    if not isiterable(wcs_list):
        msg = "Expected 'wcs_list' to be an iterable of WCS objects."
        raise ValueError(msg)

    if len(wcs_list):
        if not all(isinstance(w, gwcs.WCS) for w in wcs_list):
            msg = "All items in 'wcs_list' are to be instances of gwcs.wcs.WCS."
            raise TypeError(msg)
    else:
        msg = "'wcs_list' should not be empty."
        raise TypeError(msg)

    return True


def compute_scale(
    wcs: gwcs.wcs.WCS,
    fiducial: tuple | np.ndarray,
    disp_axis: int | None = None,
    pscale_ratio: float | None = None,
) -> float:
    """Compute the scale at the fiducial point on the detector..

    Parameters
    ----------
    wcs : ~gwcs.wcs.WCS
        Reference WCS object from which to compute a scaling factor.

    fiducial : tuple
        Input fiducial of (RA, DEC) or (RA, DEC, Wavelength) used in calculating
        reference points.

    disp_axis : int
        Dispersion axis integer. Assumes the same convention as
        ``wcsinfo.dispersion_direction``

    pscale_ratio : int
        Ratio of input to output pixel scale

    Returns
    -------
    scale : float
        Scaling factor for x and y or cross-dispersion direction.

    """
    spectral = "SPECTRAL" in wcs.output_frame.axes_type

    if spectral and disp_axis is None:
        msg = "If input WCS is spectral, a disp_axis must be given"
        raise ValueError(msg)

    crpix = np.array(wcs.invert(*fiducial))

    delta = np.zeros_like(crpix)
    spatial_idx = np.where(np.array(wcs.output_frame.axes_type) == "SPATIAL")[0]
    delta[spatial_idx[0]] = 1

    crpix_with_offsets = np.vstack((crpix, crpix + delta, crpix + np.roll(delta, 1))).T
    crval_with_offsets = wcs(*crpix_with_offsets, with_bounding_box=False)

    coords = SkyCoord(
        ra=crval_with_offsets[spatial_idx[0]],
        dec=crval_with_offsets[spatial_idx[1]],
        unit="deg",
    )
    xscale = np.abs(coords[0].separation(coords[1]).value)
    yscale = np.abs(coords[0].separation(coords[2]).value)

    if pscale_ratio is not None:
        xscale *= pscale_ratio
        yscale *= pscale_ratio

    if spectral:
        # Assuming scale doesn't change with wavelength
        # Assuming disp_axis is consistent with DataModel.meta.wcsinfo.dispersion.direction
        return float(yscale) if disp_axis == 1 else float(xscale)

    return float(np.sqrt(xscale * yscale))


def compute_fiducial(wcslist: list,
                     bounding_box: Sequence | None = None) -> np.ndarray:
    """
    Calculates the world coordinates of the fiducial point of a list of WCS objects.
    For a celestial footprint this is the center. For a spectral footprint, it is the
    beginning of its range.

    Parameters
    ----------
    wcslist : list
        A list containing all the WCS objects for which the fiducial is to be
        calculated.
    bounding_box : tuple, list, None
        The bounding box over which the WCS is valid. It can be a either tuple of tuples
        or a list of lists of size 2 where each element represents a range of
        (low, high) values. The bounding_box is in the order of the axes, axes_order.
        For two inputs and axes_order(0, 1) the bounding box can be either
        ((xlow, xhigh), (ylow, yhigh)) or [[xlow, xhigh], [ylow, yhigh]].
    
    Returns
    -------
    fiducial : np.ndarray
        A two-elements array containing the world coordinates of the fiducial point
        in the combined output coordinate frame.

    Notes
    -----
    This function assumes all WCSs have the same output coordinate frame.
    """
    axes_types = wcslist[0].output_frame.axes_type
    spatial_axes = np.array(axes_types) == "SPATIAL"
    spectral_axes = np.array(axes_types) == "SPECTRAL"
    footprints = np.hstack([w.footprint(bounding_box=bounding_box).T for w in wcslist])
    spatial_footprint = footprints[spatial_axes]
    spectral_footprint = footprints[spectral_axes]

    fiducial = np.empty(len(axes_types))
    if spatial_footprint.any():
        fiducial[spatial_axes] = _calculate_fiducial_from_spatial_footprint(spatial_footprint.T)
    if spectral_footprint.any():
        fiducial[spectral_axes] = spectral_footprint.min()
    return fiducial


def _compute_fiducial_from_footprints(footprints: list[np.ndarray]) -> tuple:
    """
    Calculates the world coordinates of the fiducial point of a list of WCS objects.
    For a celestial footprint this is the center. For a spectral footprint, it is the
    beginning of its range.

    Parameters
    ----------
    footprints : list
        A list of numpy arrays each of shape (N, 2) containing the
        (RA, Dec) vertices demarcating the footprint of the input WCSs.

    bounding_box : tuple, list, None
        The bounding box over which the WCS is valid. It can be a either tuple of tuples
        or a list of lists of size 2 where each element represents a range of
        (low, high) values. The bounding_box is in the order of the axes, axes_order.
        For two inputs and axes_order(0, 1) the bounding box can be either
        ((xlow, xhigh), (ylow, yhigh)) or [[xlow, xhigh], [ylow, yhigh]].

    Returns
    -------
    fiducial : tuple
        A tuple containing the world coordinates of the fiducial point
        in the combined output coordinate frame.

    Notes
    -----
    This function assumes all WCSs have the same output coordinate frame.
    """
    spatial_footprint = np.vstack(footprints)
    return _calculate_fiducial_from_spatial_footprint(spatial_footprint)


def calc_rotation_matrix(roll_ref: float, v3i_yangle: float, vparity: int = 1) -> list[float]:
    r"""Calculate the rotation matrix.

    Parameters
    ----------
    roll_ref : float
        Telescope roll angle of V3 North over East at the ref. point in radians

    v3i_yangle : float
        The angle between ideal Y-axis and V3 in radians.

    vparity : int
        The x-axis parity, usually taken from the JWST SIAF parameter VIdlParity.
        Value should be "1" or "-1".

    Returns
    -------
    matrix: list
        A list containing the rotation matrix elements in column order.

    Notes
    -----
    The rotation matrix is

    .. math::
        PC = \\begin{bmatrix}
                pc_{1,1} & pc_{2,1} \\\\
                pc_{1,2} & pc_{2,2}
            \\end{bmatrix}
    """
    if vparity not in (1, -1):
        msg = f"vparity should be 1 or -1. Input was: {vparity}"
        raise ValueError(msg)

    rel_angle = roll_ref - (vparity * v3i_yangle)

    pc1_1 = vparity * np.cos(rel_angle)
    pc1_2 = np.sin(rel_angle)
    pc2_1 = vparity * -np.sin(rel_angle)
    pc2_2 = np.cos(rel_angle)

    return [pc1_1, pc1_2, pc2_1, pc2_2]


def wcs_from_footprints(
    wcs_list: list[gwcs.wcs.WCS],
    ref_wcs: gwcs.wcs.WCS,
    ref_wcsinfo: dict,
    transform: astropy.modeling.models.Model | None = None,
    bounding_box: Sequence | None = None,
    pscale_ratio: float | None = None,
    pscale: float | None = None,
    rotation: float | None = None,
    shape: Sequence | None = None,
    crpix: Sequence | None = None,
    crval: Sequence | None = None,
) -> gwcs.wcs.WCS:
    """
    Create a WCS from a list of input datamodels.

    A fiducial point in the output coordinate frame is created from  the
    footprints of all WCS objects. For a spatial frame this is the center
    of the union of the footprints. For a spectral frame the fiducial is in
    the beginning of the footprint range.
    If ``refmodel`` is None, the first WCS object in the list is considered
    a reference. The output coordinate frame and projection (for celestial frames)
    is taken from ``refmodel``.
    If ``transform`` is not supplied, a compound transform is created using
    CDELTs and PC.
    If ``bounding_box`` is not supplied, the `bounding_box` of the new WCS is computed
    from `bounding_box` of all input WCSs.

    Parameters
    ----------
    wcs_list : list
        A list of valid datamodels.

    ref_wcs :
        A valid datamodel whose WCS is used as reference for the creation of the output
        coordinate frame, projection, and scaling and rotation transforms.
        If not supplied the first model in the list is used as ``refmodel``.

    ref_wcsinfo : dict
        A dictionary containing the WCS FITS keywords and corresponding values.

    transform : ~astropy.modeling.Model
        A transform, passed to :py:func:`gwcs.wcstools.wcs_from_fiducial`
        If not supplied `Scaling | Rotation` is computed from ``refmodel``.

    bounding_box : tuple
        Bounding_box of the new WCS.
        If not supplied it is computed from the bounding_box of all inputs.

    pscale_ratio : float, None
        Ratio of input to output pixel scale. Ignored when either
        ``transform`` or ``pscale`` are provided.

    pscale : float, None
        Absolute pixel scale in degrees. When provided, overrides
        ``pscale_ratio``. Ignored when ``transform`` is provided.

    rotation : float, None
        Position angle of output image's Y-axis relative to North.
        A value of 0.0 would orient the final output image to be North up.
        The default of `None` specifies that the images will not be rotated,
        but will instead be resampled in the default orientation for the camera
        with the x and y axes of the resampled image corresponding
        approximately to the detector axes. Ignored when ``transform`` is
        provided.

    shape : tuple of int, None
        Shape of the image (data array) using ``np.ndarray`` convention
        (``ny`` first and ``nx`` second). This value will be assigned to
        ``pixel_shape`` and ``array_shape`` properties of the returned
        WCS object.

    crpix : tuple of float, None
        Position of the reference pixel in the image array.  If ``crpix`` is not
        specified, it will be set to the center of the bounding box of the
        returned WCS object.

    crval : tuple of float, None
        Right ascension and declination of the reference pixel. Automatically
        computed if not provided.

    wcs_list : list
        A list of WCS objects. If not supplied, the WCS objects are extracted
        from the input datamodels.

    Returns
    -------
    wcs_new : ~gwcs.wcs.WCS
        The WCS object corresponding to the combined input footprints.

    """
    msg = ("wcs_from_footprints is deprecated and will be removed in a future release."
           "It is recommended to use wcs_from_sregions instead.")
    warnings.warn(msg, DeprecationWarning, stacklevel=2)
    _validate_wcs_list(wcs_list)
    footprints = [w.footprint() for w in wcs_list]
    if ref_wcs is None:
        ref_wcs = wcs_list[0]
    return wcs_from_sregions(
        footprints,
        ref_wcs,
        ref_wcsinfo,
        transform=transform,
        pscale_ratio=pscale_ratio,
        pscale=pscale,
        rotation=rotation,
        shape=shape,
        crpix=crpix,
        crval=crval,
    )


def _sregion_to_footprint(s_region: str) -> np.ndarray:
    """
    Parameters
    ----------
    s_region : str
        The S_REGION header keyword

    Returns
    -------
    footprint : np.array
        A 2D array of the footprint of the region, shape (N, 2)
    """
    no_prefix = re.sub(r"[a-zA-Z]", "", s_region)
    return np.array(no_prefix.split(), dtype=float).reshape(-1, 2)


def wcs_from_sregions(
    footprints: list[np.ndarray] | list[str],
    ref_wcs: gwcs.wcs.WCS,
    ref_wcsinfo: dict,
    transform: astropy.modeling.models.Model | None = None,
    pscale_ratio: float | None = None,
    pscale: float | None = None,
    rotation: float | None = None,
    shape: Sequence | None = None,
    crpix: Sequence | None = None,
    crval: Sequence | None = None,
) -> gwcs.wcs.WCS:
    """
    Create a WCS from a list of input s_regions or footprints.

    A fiducial point in the output coordinate frame is created from  the
    footprints of all WCS objects. For a spatial frame this is the center
    of the union of the footprints. For a spectral frame the fiducial is in
    the beginning of the footprint range.
    If ``refmodel`` is None, the first WCS object in the list is considered
    a reference. The output coordinate frame and projection (for celestial frames)
    is taken from ``refmodel``.
    If ``transform`` is not supplied, a compound transform is created using
    CDELTs and PC.
    If ``bounding_box`` is not supplied, the `bounding_box` of the new WCS is computed
    from `bounding_box` of all input WCSs.

    Parameters
    ----------
    footprints : list of np.ndarray or list of str
        If list elements are numpy arrays, each should have shape (N, 2) and contain
        (RA, Dec) vertices demarcating the footprint of the input WCSs.
        If list elements are strings, each should be the S_REGION header keyword 
        containing (RA, Dec) vertices demarcating the footprint of the input WCSs.

    ref_wcs :
        A WCS used as reference for the creation of the output
        coordinate frame, projection, and scaling and rotation transforms.

    ref_wcsinfo : dict
        A dictionary containing the WCS FITS keywords and corresponding values.

    transform : ~astropy.modeling.Model
        A transform, passed to :py:func:`gwcs.wcstools.wcs_from_fiducial`
        If not supplied `Scaling | Rotation` is computed from ``refmodel``.

    pscale_ratio : float, None
        Ratio of input to output pixel scale. Ignored when either
        ``transform`` or ``pscale`` are provided.

    pscale : float, None
        Absolute pixel scale in degrees. When provided, overrides
        ``pscale_ratio``. Ignored when ``transform`` is provided.

    rotation : float, None
        Position angle of output image's Y-axis relative to North.
        A value of 0.0 would orient the final output image to be North up.
        The default of `None` specifies that the images will not be rotated,
        but will instead be resampled in the default orientation for the camera
        with the x and y axes of the resampled image corresponding
        approximately to the detector axes. Ignored when ``transform`` is
        provided.

    shape : tuple of int, None
        Shape of the image (data array) using ``np.ndarray`` convention
        (``ny`` first and ``nx`` second). This value will be assigned to
        ``pixel_shape`` and ``array_shape`` properties of the returned
        WCS object.

    crpix : tuple of float, None
        Position of the reference pixel in the image array.  If ``crpix`` is not
        specified, it will be set to the center of the bounding box of the
        returned WCS object.

    crval : tuple of float, None
        Right ascension and declination of the reference pixel. Automatically
        computed if not provided.

    Returns
    -------
    wcs_new : ~gwcs.wcs.WCS
        The WCS object corresponding to the combined input footprints.

    """
    footprints = [_sregion_to_footprint(s_region) 
                  if isinstance(s_region, str) else s_region 
                  for s_region in footprints]
    fiducial = _calculate_fiducial(footprints, crval=crval)

    transform = _generate_tranform(
        ref_wcs,
        wcsinfo=ref_wcsinfo,
        pscale_ratio=pscale_ratio,
        pscale=pscale,
        rotation=rotation,
        ref_fiducial=np.array([ref_wcsinfo["ra_ref"], ref_wcsinfo["dec_ref"]]),
        transform=transform,
    )

    return _calculate_new_wcs(
        wcs=ref_wcs,
        shape=shape,
        crpix=crpix,
        footprints=footprints,
        fiducial=fiducial,
        transform=transform,
    )


def compute_s_region_imaging(wcs: gwcs.wcs.WCS,
                             shape: Sequence | None = None,
                             center: bool = True) -> str | None:
    """
    Update the ``S_REGION`` keyword using the WCS footprint.

    Parameters
    ----------
    wcs : ~gwcs.wcs.WCS
        The WCS object.

    shape : tuple, optional
        Shape of input model data array. Used to compute the bounding box if not
        provided in the WCS object, and required in that case. The default is None.

    center : bool, optional
        Whether or not to use the center of the pixel as reference for the
        coordinates, by default True

    Returns
    -------
    s_region : str
        String containing the S_REGION object.
    """
    bbox = wcs.bounding_box
    if shape is None and bbox is None:
        msg = "If wcs.bounding_box is not specified, shape must be provided."
        raise ValueError(msg)

    if shape is not None and bbox is None:
        bbox = wcs_bbox_from_shape(shape)
        wcs.bounding_box = bbox

    # footprint is an array of shape (2, 4) as we
    # are interested only in the footprint on the sky
    ### TODO: we shouldn't use center=True in the call below because we want to
    ### calculate the coordinates of the footprint based on the *bounding box*,
    ### which means we are interested in each pixel's vertice, not its center.
    ### By using center=True, a difference of 0.5 pixel should be accounted for
    ### when comparing the world coordinates of the bounding box and the footprint.
    footprint = wcs.footprint(bbox, center=center, axis_type="spatial").T
    # take only imaging footprint
    footprint = footprint[:2, :]

    # Make sure RA values are all positive
    negative_ind = footprint[0] < 0
    if negative_ind.any():
        footprint[0][negative_ind] = 360 + footprint[0][negative_ind]

    footprint = footprint.T
    return compute_s_region_keyword(footprint)


def wcs_bbox_from_shape(shape: Sequence) -> tuple:
    """Create a bounding box from the shape of the data.

    This is appropriate to attach to a wcs object

    Parameters
    ----------
    shape : tuple
        The shape attribute from a `np.ndarray` array

    Returns
    -------
    bbox : tuple
        Bounding box in x, y order.
    """
    return (-0.5, shape[-1] - 0.5), (-0.5, shape[-2] - 0.5)


def compute_s_region_keyword(footprint: np.ndarray) -> str | None:
    """Update the S_REGION keyword.

    Parameters
    ----------
    footprint :
        A 4x2 numpy array containing the coordinates of the vertices of the footprint.

    Returns
    -------
    s_region : str
        String containing the S_REGION object.
    """
    s_region = "POLYGON ICRS  {:.9f} {:.9f} {:.9f} {:.9f} {:.9f} {:.9f} {:.9f} {:.9f}".format(
       *footprint.flatten()
    )
    if "nan" in s_region:
        # do not update s_region if there are NaNs.
        log.info("There are NaNs in s_region, S_REGION not updated.")
        return None
    log.info("Update S_REGION to %s", s_region)
    return s_region


def reproject(wcs1: gwcs.wcs.WCS, wcs2: gwcs.wcs.WCS) -> Callable:
    """
    Given two WCSs or transforms return a function which takes pixel
    coordinates in the first WCS or transform and computes them in pixel coordinates
    in the second one. It performs the forward transformation of ``wcs1`` followed by the
    inverse of ``wcs2``.

    Parameters
    ----------
    wcs1 : astropy.wcs.WCS or gwcs.wcs.WCS
        Input WCS objects or transforms.
    wcs2 : astropy.wcs.WCS or gwcs.wcs.WCS
        Output WCS objects or transforms.

    Returns
    -------
        Function to compute the transformations.  It takes x, y
        positions in ``wcs1`` and returns x, y positions in ``wcs2``.
    """

    def _get_forward_transform_func(wcs1):
        """Get the forward transform function from the input WCS. If the wcs is a
        fitswcs.WCS object all_pix2world requires three inputs, the x (str, ndarrray),
        y (str, ndarray), and origin (int). The origin should be between 0, and 1
        https://docs.astropy.org/en/latest/wcs/index.html#loading-wcs-information-from-a-fits-file
        ).
        """
        if isinstance(wcs1, fitswcs.WCS):
            forward_transform = wcs1.all_pix2world
        elif isinstance(wcs1, gwcs.WCS):
            forward_transform = wcs1.forward_transform
        else:
            msg = "Expected input to be astropy.wcs.WCS or gwcs.WCS object"
            raise TypeError(msg)
        return forward_transform

    def _get_backward_transform_func(wcs2):
        if isinstance(wcs2, fitswcs.WCS):
            backward_transform = wcs2.all_world2pix
        elif isinstance(wcs2, gwcs.WCS):
            backward_transform = wcs2.backward_transform
        else:
            msg = "Expected input to be astropy.wcs.WCS or gwcs.WCS object"
            raise TypeError(msg)
        return backward_transform

    def _reproject(x: float | np.ndarray, y: float | np.ndarray) -> tuple:
        """
        Reprojects the input coordinates from one WCS to another.

        Parameters
        ----------
        x : float or np.ndarray
            x-coordinate(s) to be reprojected.
        y : float or np.ndarray
            y-coordinate(s) to be reprojected.

        Returns
        -------
        tuple
            Tuple of np.ndarrays including reprojected x and y coordinates.
        """
        # example inputs to resulting function (12, 13, 0) # third number is origin
        # uses np.arrays for shape functionality
        if not isinstance(x, (np.ndarray)):
            x = np.array(x)
        if not isinstance(y, (np.ndarray)):
            y = np.array(y)
        if x.shape != y.shape:
            msg = "x and y must be the same length"
            raise ValueError(msg)
        sky = _get_forward_transform_func(wcs1)(x, y, 0)

        # rearrange into array including flattened x and y values
        flat_sky = [axis.flatten() for axis in sky]
        det = np.array(_get_backward_transform_func(wcs2)(flat_sky[0], flat_sky[1], 0))
        det_reshaped = [axis.reshape(x.shape) for axis in det]

        return tuple(det_reshaped)

    return _reproject


import os
import glob
from multiprocessing import Pool, cpu_count
from spike import psfgen, tools
from astropy.io import fits
from astropy.wcs import WCS, utils
import numpy as np
import subprocess
from subprocess import call
import warnings

def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
	return '%s:%s: %s: %s\n' % (filename, lineno, category.__name__, message)

warnings.formatwarning = warning_on_one_line

##########
# * * * *
##########


def hst(img_dir, obj, img_type, inst, camera = None, method='TinyTim', usermethod = None, 
		savedir = 'psfs', drizzleimgs = False, pretweaked = False,
		keeporig = True, plot = False, verbose = False, parallel = False, out = 'fits', 
		tweakparams = {'threshold':6.0, 
					   'searchrad':3.0, 
					   'dqbits':-16, 
					   'configobj':None, 
					   'interactive':False, 
					   'shiftfile':True, 
					   'expand_refcat':True,
					   'outshifts':'shift_searchrad.txt', 
					   'updatehdr':True,
					   'wcsname': 'TWEAK'}, 
		drizzleparams = {'preserve':False,
						 'driz_cr_corr':False,
						 'clean':False,
						 'configobj':None,
						 'final_pixfrac':0.8,
						 'build':True,
						 'combine_type':'imedian',
						 'static':False},
		returnpsf = 'full',
		cutout_fov = 151,
		savecutout = True,
		finalonly = False,
		removedir = 'toremove',
		**kwargs):
	"""
	Generate drizzled HST PSFs.

	Parameters:
		img_dir (str): Path to directory containing calibrated files for which model PSF will be generated.
			If using the tweakreg step, best to include a drizzled file, as well, which can be used as a reference.
		obj(str, arr-like): Name or coordinates of object of interest in HH:MM:DD DD:MM:SS or degree format.
		img_type (str): e.g, 'flc', 'flt', 'cal', 'c0m' -- specifies which file-type to include.
			spike currently only works with MEF files (since astrodrizzle only works with MEF files).
		inst (str): 'ACS', 'WFC3', 'WFPC', 'WFPC2', NICMOS'
		camera (str): 'WFC', 'HRC' (ACS), 'UVIS', 'IR' (WFC3) -- MUST BE SPECIFIED FOR ACS, WFC
		method (str): 'TinyTim', 'TinyTim_Gillis', 'STDPSF' (empirical),
				'epsf' (empirical), 'PSFEx' (empirical) -- see spike.psfgen for details -- or 'USER';
				if 'USER', usermethod should be a function that generates, or path to a directory of user-generated, PSFs 
				named [imgprefix]_[coords]_[band]_topsf.fits, e.g., imgprefix_23.31+30.12_F814W_topsf.fits or 
				imgprefix_195.78-46.52_F555W_topsf.fits
		usermethod (func or str): If method = 'USER', usermethod should be a function that generates, or path to a 
				directory of user-generated, PSFs named [imgprefix]_[coords]_[band]_psf.fits, e.g., 
				imgprefix_23.31+30.12_F814W_psf.fits or imgprefix_195.78-46.52_F555W_psf.fits, where the 
				imgprefix corresponds to the name of the relevant flt/flc/c0f/c1f/... files in the directory and the 
				headers are from the original images (see spike.tools.rewrite_fits, which can be used to this end).
		savedir (str): Where the PSF models and drizzled PSF will be saved. Defaults to 'psfs'.
		drizzleimgs (bool): If True, will drizzle the input images at the same time as creating a drizzled psf.
		pretweaked (bool): If True, skips TweakReg steps to include fine WCS corrections.
		keeporig (bool): If True (and pretweaked = False), create copy of img_dir before TweakReg.
		plot (bool): If True, saves .pngs of the model PSFs.
		verbose (bool): If True, prints progress messages.
		parallel (bool): If True, runs PSF generation in parallel.
		out (str): 'fits' or 'asdf'. Output for the drizzled PSF. If 'asdf', .asdf AND .fits are saved.
		tweakparams (dict): Dictionary of keyword arguments for drizzlepac.tweakreg. See the drizzlepac documentation
			for a full list.
		drizzleparams (dict): Dictionary of keyword arguments for drizzlepac.astrodrizzle. See the drizzlepac 
			documentation for a full list.
		returnpsf (str): 'full', 'crop', or None. If None, spike.psf.hst does not return anything. If 'full' (default),
			returns the PSF in the full spatial context of the processed image. If 'crop', returns the region immediately
			around the PSF (size of cutout set by cutout_fov).
		cutout_fov (int): Side length in pixels of square cutout region centered on PSF. Used if returnpsf = 'crop'.
		savecutout (bool): If True, save a .fits file with the cutout region, including WCS. Only used if returnpsf = 'crop'.
		finalonly (bool): If True, only retains final drizzled/resampled data products in savedir and deletes intermediate products.
		removedir (str): Directory (**to be deleted**) that stores intermediate products for removal. Default is 'toremove'.
		**kwargs: Keyword arguments for PSF generation function.

	Returns:
		Generates model PSFs and drizzled PSF. (If drizzledimgs = True, also produces drizzled image from input files.)

		If returnpsf = 'full', will return each of the full drizzled PSF images in an object, filter indexed dict.
		If returnpsf = 'crop', will return a cutout region of the drizzled PSF images (around the PSF) in an obj, filt indexed dict.
	"""
	from drizzlepac import tweakreg, tweakback, astrodrizzle

	if img_type.lower() in ['drc', 'drz']:
		raise Exception('%s files are already drizzled. spike works with calibrated, but not-yet-combined images -- e.g., flc, crf, cal.'%img_type)

	if img_dir[:-1] != '/':
		img_dir += '/' #force paths to work out

	if keeporig and not pretweaked:
		if not os.path.exists(img_dir+'_orig'):
			os.makedirs(img_dir+'_orig')
		os.system('cp -r '+img_dir+'*_'+img_type+'.fits '+img_dir+'_orig')
		if verbose:
			print('Made copy of '+img_dir)


	imgs = sorted(glob.glob(img_dir+'*'+img_type+'.fits'))

	if inst.upper() in ['ACS', 'WFC3']:
		imcam = inst.upper()+'/'+camera.upper()
	if inst.upper() in ['WFPC1', 'WFPC2', 'NICMOS', 'STIS']:
		imcam = inst.upper()

	if inst.upper() == 'WFPC2':
		updatewcs = True

	genpsf = True
	if method.upper() not in ['TINYTIM', 'TINYTIM_GILLIS', 'STDPSF', 'EPSF', 'PSFEX', 'USER']:
		raise Exception('method must be one of TINYTIM, TINYTIM_GILLIS, STDPSF, EPSF, PSFEX, USER')
	if method.upper() == 'TINYTIM':
		if inst.upper() == 'WFC3':
			warnings.warn('TinyTim is not recommended for modeling WFC3 PSFs. See https://www.stsci.edu/hst/instrumentation/focus-and-pointing/focus/tiny-tim-hst-psf-modeling.',
				Warning, stacklevel = 2)
		psffunc = psfgen.tinypsf
	if method.upper() == 'TINYTIM_GILLIS':
		if (inst.upper() != 'ACS') and (camera.upper() != 'WFC'):
			warnings.warn('The Gillis (2019) code is made for/tested on ACS/WFC and no modification is made here to generalize it to other HST instruments/cameras.')
		psffunc = psfgen.tinygillispsf
	if method.upper() == 'STDPSF':
		if inst.upper() in ['WFPC', 'WFPC1']:
			raise ValueError("There is no available STDPSF grid for WFPC imaging. Please select a different PSF generation method.")
		psffunc = psfgen.stdpsf
	if method.upper() == 'EPSF':
		psffunc = psfgen.effpsf
	if method.upper() == 'PSFEX':
		psffunc = psfgen.psfex
	if method.upper() == 'USER':
		if type(usermethod) == str: #check if user input is path to directory
			genpsf = False
		if type(usermethod) != str: #or function
			psffunc = method

	filelist = {} # generate list of files to tweak -- by filter
	for fi in imgs:
		hdu = fits.open(fi)
		try: #get filter
			filt = hdu[0].header['FILTER']
		except:
			if hdu[0].header['FILTER1'].startswith('F'):
				filt = hdu[0].header['FILTER1']
			else:
				filt = hdu[0].header['FILTER2']
		if filt not in filelist.keys():
			filelist[filt] = []
		filelist[filt].append(fi)

	if not pretweaked:
		# note that if there are many input files, tweakreg will be very slow and prone
		# to overuse of RAM	
		for fk in filelist.keys():
			tweakreg.TweakReg(filelist[fk], **tweakparams)

	drizzlelist = {} #write file prefixes to drizzle per object per filter
	if genpsf: #generate model PSFs for each image + object
		if type(obj) == str: #check number of objects
			drizzlelist[obj] = {}
			skycoords = tools.objloc(obj)
			for i in imgs:
				pos = tools.checkpixloc(skycoords, i, inst, camera)

				coordstring = str(skycoords.ra)
				if skycoords.dec.deg >= 0:
					coordstring += '+'+str(skycoords.dec)
				if skycoords.dec.deg < 0:
					coordstring += str(skycoords.dec)

				modname = i.replace('.fits', '_'+coordstring+'_%s'%pos[3]+'_topsf.fits')
				if np.isfinite(pos[0]): #confirm object falls onto image
					if pos[3] not in drizzlelist[obj].keys():
						drizzlelist[obj][pos[3]] = []
					drizzlelist[obj][pos[3]].append(modname)

					psffunc(skycoords, i, imcam, pos, plot, verbose, **kwargs)

		if type(obj) != str: #if multiple objects, option to parallelize 
			skycoords = [] #only open each FITS file once

			for o in obj:
				drizzlelist[o] = {}
				skycoords.append(tools.objloc(o))
			
			for i in imgs:

				if parallel:
					if method.upper() == 'PSFEX':
						warnings.warn('Warning: Check your config and param files to ensure output files have unique names.', Warning, stacklevel = 2)
					pool = Pool(processes=(cpu_count() - 1))
					for j, coord in enumerate(skycoords):

						pos = tools.checkpixloc(coord, i, inst, camera)

						coordstring = str(coord.ra)
						if coord.dec.deg >= 0:
							coordstring += '+'+str(coord.dec)
						if coord.dec.deg < 0:
							coordstring += str(coord.dec)

						modname = i.replace('.fits', '_'+coordstring+'_%s'%pos[3]+'_topsf.fits')
						if np.isfinite(pos[0]): #confirm that object falls onto detector
							if pos[3] not in drizzlelist[obj[j]].keys():
								drizzlelist[obj[j]][pos[3]] = []
							drizzlelist[obj[j]][pos[3]].append(modname)

							pool.apply_async(psffunc, args = (coord, i, imcam, pos, plot, verbose), kwds = kwargs)
					pool.close()
					pool.join()

				if not parallel:
					for j, coord in enumerate(skycoords):
						pos = tools.checkpixloc(coord, i, inst, camera)

						coordstring = str(coord.ra)
						if coord.dec.deg >= 0:
							coordstring += '+'+str(coord.dec)
						if coord.dec.deg < 0:
							coordstring += str(coord.dec)

						modname = i.replace('.fits', '_'+coordstring+'_%s'%pos[3]+'_topsf.fits')
						if np.isfinite(pos[0]): #confirm that object falls onto detector
							if pos[3] not in drizzlelist[obj[j]].keys():
								drizzlelist[obj[j]][pos[3]] = []
							drizzlelist[obj[j]][pos[3]].append(modname)

						psffunc(coord, i, imcam, pos, plot, verbose, **kwargs) 
					
	if not genpsf:
		userpsfs = sorted(glob.glob(usermethod))

		for up in userpsfs:
			im, obj, filt, _ = up.split('_')
			if obj not in drizzlelist.keys():
				drizzlelist[obj] = {}
			if filt not in drizzlelist[obj].keys():
				drizzlelist[obj][filt] = []
			drizzlelist[obj][filt].append(im)
			

	if keeporig:
		drizzleparams['preserve'] = True #reset parameter to ensure that original files maintained

	for do in drizzlelist.keys():
		cstring = tools.objloc(do)
		coordstring = str(cstring.ra)
		if cstring.dec.deg >= 0:
			coordstring += '+'+str(cstring.dec)
		if cstring.dec.deg < 0:
			coordstring += str(cstring.dec)

		if parallel:
			pool = Pool(processes=(cpu_count() - 1))
			for dk in drizzlelist[do].keys():
				outname = coordstring+'_'+dk+'_psf'
				drizzleparams['output'] = img_dir + outname #set output based on coord, filter
				pool.apply_async(astrodrizzle.AstroDrizzle, args = (drizzlelist[do][dk]), kwds = drizzleparams)
			pool.close()
			pool.join()
		if not parallel:
			for dk in drizzlelist[do].keys():
				outname = coordstring+'_'+dk+'_psf'
				drizzleparams['output'] = img_dir + outname #set output based on coord, filter
				astrodrizzle.AstroDrizzle(drizzlelist[do][dk], **drizzleparams)

	drzs = np.concatenate((sorted(glob.glob('%s*_drc.fits'%img_dir)), 
		sorted(glob.glob('%s*_drz.fits'%img_dir)), sorted(glob.glob('%s*_mos.fits'%img_dir))))

	if len(drzs) == 0:
		raise Exception('No co-added/resampled output files created. Check your input path, coordinates and the output of the PSF generation steps.')

	for dr in drzs: #rename drizzled outputs to something more manageable
		flist = dr.split('_')
		suff_ = flist[-1].split('.')[0]
		filt_ = flist[1]
		obj_ = flist[0]

		os.system('mv %s %s%s_%s_psf_%s.fits'%(dr, img_dir, obj_, filt_, suff_))

	suff = suff_ # store suffix, as there should be no variation within one run

	
	if drizzleimgs: # useful for processing all images + PSFs simultaneously
		drizzleparams['driz_cr_corr'] = True #reset parameters turned off for PSF
		drizzleparams['static'] = True
		for fk in filelist.keys():
			drizzleparams['output'] = img_dir + '%s_img'%fk #set output name with filter
			astrodrizzle.AstroDrizzle(filelist[fk], **drizzleparams)

	if not finalonly:
		# clean up step to move all of the PSF files to the relevant directory
		# should grab all .pngs, .fits etc.
		if not os.path.exists(savedir):
			os.makedirs(savedir)

		os.system('mv %s*_psf* %s'%(img_dir, savedir)) # generated PSF models
		os.system('mv %s*.psf %s'%(img_dir, savedir))
		os.system('mv %s*_topsf* %s'%(img_dir, savedir)) # tweaked and drizzled PSF models

		## clean up other files generated in the process
		os.system('mv %s*.cat %s'%(img_dir, savedir))
		os.system('mv %s*_mask.fits %s'%(img_dir, savedir))
		os.system('mv *_sci1.fits %s'%(savedir))


		if verbose:
			print('Moved PSF files to %s'%savedir)


	if finalonly:
		# clean up step to move all of the resampled files to the relevant directory
		if not os.path.exists(savedir):
			os.makedirs(savedir)

		os.system('mv %s*_drz* %s'%(img_dir, savedir)) #move files to preserve
		os.system('mv %s*_drc* %s'%(img_dir, savedir))
		os.system('mv %s*_mos* %s'%(img_dir, savedir))

		if verbose:
			print('Moved drizzled files to %s'%savedir)

		if os.path.exists(removedir):
			# raise warning if removedir exists; not an error, though since some people might create it for this purpose
			warnings.warn('%s already exists. This directory and its contents will be deleted.'%(removedir),
				Warning, stacklevel = 2)

		if not os.path.exists(removedir): #directory to remove excess files
			os.makedirs(removedir)

		os.system('mv %s*_psf* %s'%(img_dir, removedir)) # generated PSF models
		os.system('mv %s*.psf %s'%(img_dir, removedir))
		os.system('mv %s*_topsf* %s'%(img_dir, removedir))

		## clean up other files generated in the process
		os.system('mv %s*.cat %s'%(img_dir, removedir))
		os.system('mv %s*_mask.fits %s'%(img_dir, removedir))
		os.system('mv *_sci1.fits %s'%(removedir))

		os.system('rm -r %s'%s(removedir))

		if verbose:
			print('Deleted intermediate products and removedir.')




	if out == 'asdf':
		# .asdf file read out in addition to .fits
		if savedir.split('/')[-1] != '':
			savedir += '/'

		sufs = ['drc', 'drz', 'mos']
		dout = np.concatenate((sorted(glob.glob(savedir+'*_drc.fits')), 
			sorted(glob.glob(savedir+'*_drz.fits')), sorted(glob.glob(savedir+'*_mos.fits'))))
		for di in dout:
			tools.to_asdf(di)

		if verbose:
			print('Generated ASDF output')

	if returnpsf:
		returndict = {}
		for o, do in enumerate(drizzlelist.keys()):
			if type(obj) == str:
				coordstring = str(skycoords.ra)
				if skycoords.dec.deg >= 0:
					coordstring += '+'+str(skycoords.dec)
				if skycoords.dec.deg < 0:
					coordstring += str(skycoords.dec)

			if type(obj) != str:
				coordstring = str(skycoords[o].ra)
				if skycoords[o].dec.deg >= 0:
					coordstring += '+'+str(skycoords[o].dec)
				if skycoords[o].dec.deg < 0:
					coordstring += str(skycoords[o].dec)

			returndict[do] = {}
			for dk in drizzlelist[do].keys():

				if savedir.split('/')[-1] != '':
					savedir += '/'

				if returnpsf == 'full':	
					dr_psf = fits.open(savedir+'%s_%s_psf_%s.fits'%(coordstring, dk, suff))
					returndict[do][dk] = dr_psf[1].data

				if returnpsf == 'crop':
					crop = tools.cutout(img = savedir+'%s_%s_psf_%s.fits'%(coordstring, dk, suff), 
									coords = tools.objloc(do), fov_pixel = cutout_fov, save = savecutout)
					returndict[do][dk] = crop

		return returndict



def jwst(img_dir, obj, inst, img_type = 'cal', camera = None, method = 'WebbPSF', usermethod = None, 
		savedir = 'psfs', drizzleimgs = False, pretweaked = False, usecrds = False, 
		keeporig = True, plot = False, verbose = False, parallel = False, out = 'fits',
		returnpsf = 'full', cutout_fov = 151, savecutout = True, finalonly = False, 
		removedir = 'toremove', tweakparams = {}, drizzleparams = {'allowed_memory':0.5}, **kwargs):
	"""
	Generate drizzled James Webb Space Telescope PSFs.

	Parameters:
		img_dir (str): Path to directory containing calibrated files for which model PSF will be generated.
				If using the tweakreg step, best to include a drizzled file, as well, which can be used as a reference.
		obj(str, arr-like): Name or coordinates of object of interest in HH:MM:DD DD:MM:SS or degree format.
		img_type (str): e.g, 'cal', 'calints', 'crf', 'crfints' -- specifies which file-type to include.
		inst (str): 'MIRI', 'NIRCAM', 'NIRISS'
		camera (str): 'Imaging', 'AMI' -- MUST BE SPECIFIED FOR NIRISS
		method (str): 'WebbPSF', 'STDPSF' (empirical), 'epsf' (empirical), 'PSFEx' (empirical) -- see spike.psfgen for details -- or 'USER';
				if 'USER', usermethod should be a function that generates, or path to a directory of user-generated, PSFs 
				named [imgprefix]_[coords]_[band]_psf.fits, e.g., imgprefix_23.31+30.12_F814W_psf.fits or 
				imgprefix_195.78-46.52_F555W_psf.fits. Note: the WebbPSF name is maintained here in lieu of STPSF to avoid
				confusion with the generation of empirical STDPSFs.
		usermethod (func or str): If method = 'USER', usermethod should be a function that generates, or path to a 
				directory of user-generated, PSFs named [imgprefix]_[coords]_[band]_psf.fits, e.g., 
				imgprefix_23.31+30.12_F814W_psf.fits or imgprefix_195.78-46.52_F555W_psf.fits, where the 
				imgprefix corresponds to the name of the relevant flt/flc/c0f/c1f/... files in the directory and the 
				headers are from the original images (see spike.tools.rewrite_fits, which can be used to this end).
		savedir (str): Where the PSF models and drizzled PSF will be saved. Defaults to 'psfs'.
		drizzleimgs (bool): If True, will drizzle the input images at the same time as creating a drizzled psf.
		pretweaked (bool): If True, skips tweak step to include fine WCS corrections.
		usecrds (bool): If True, use CRDS config settings as defaults. 
		keeporig (bool): If True (and pretweaked = False), create copy of img_dir before tweak.
		plot (bool): If True, saves .pngs of the model PSFs.
		verbose (bool): If True, prints progress messages.
		parallel (bool): If True, runs PSF generation in parallel.
		out (str): 'fits' or 'asdf'. Output for the drizzled PSF. If 'asdf', .asdf AND .fits are saved.
		returnpsf (str): 'full', 'crop', or None. If None, spike.psf.jwst does not return anything. If 'full' (default),
			returns the PSF in the full spatial context of the processed image. If 'crop', returns the region immediately
			around the PSF (size of cutout set by cutout_fov).
		cutout_fov (int): Side length in pixels of square cutout region centered on PSF. Used if returnpsf = 'crop'.
		savecutout (bool): If True, save a .fits file with the cutout region, including WCS. Only used if returnpsf = 'crop'.
		finalonly (bool): If True, only retains final drizzled/resampled data products in savedir and deletes intermediate products.
		removedir (str): Directory (**to be deleted**) that stores intermediate products for removal. Default is 'toremove'.
		tweakparams (dict): Dictionary of keyword arguments for drizzlepac.tweakreg. See the drizzlepac documentation
				for a full list. See here: https://jwst-pipeline.readthedocs.io/en/latest/jwst/tweakreg/README.html#step-arguments
		drizzleparams (dict): Dictionary of keyword arguments for drizzlepac.astrodrizzle. See the drizzlepac 
				documentation for a full list.
		**kwargs: Keyword arguments for PSF generation function.

	Returns:
		Generates model PSFs and drizzled PSF. (If drizzledimgs = True, also produces drizzled image from input files.)

		If returnpsf = 'full', will return each of the full drizzled PSF images in an object, filter indexed dict.
		If returnpsf = 'crop', will return a cutout region of the drizzled PSF images (around the PSF) in an obj, filt indexed dict.
	"""

	os.environ['CRDS_SERVER_URL']="https://jwst-crds.stsci.edu"

	if not usecrds:
		os.environ["STPIPE_DISABLE_CRDS_STEPPARS"] = 'True'

	from spike.jwstcal import resample_step
	from spike.jwstcal import tweakreg_tweakreg_step as tweakreg_step

	if img_dir[:-1] != '/':
		img_dir += '/' #force paths to work out

	if keeporig and not pretweaked:
		if not os.path.exists(img_dir+'_orig'):
			os.makedirs(img_dir+'_orig')
		os.system('cp -r '+img_dir+'*_'+img_type+'.fits '+'img_dir'+'_orig')
		if verbose:
			print('Made copy of '+img_dir)


	imgs = sorted(glob.glob(img_dir+'*'+img_type+'.fits'))

	imcam = inst.upper()

	genpsf = True
	if method.upper() not in ['WEBBPSF', 'STDPSF', 'EPSF', 'PSFEX', 'USER']:
		raise Exception('tool must be one of WEBBPSF, STDPSF, EPSF, PSFEX, USER')
	if method.upper() == 'WEBBPSF':
		psffunc = psfgen.jwpsf
	if method.upper() == 'STDPSF':
		psffunc = psfgen.stdpsf
	if method.upper() == 'EPSF':
		psffunc = psfgen.effpsf
	if method.upper() == 'PSFEX':
		psffunc = psfgen.psfex
	if method.upper() == 'USER':
		if type(usermethod) == str: #check if user input is path to directory
			genpsf = False
		if type(usermethod) != str: #or function
			psffunc = method

	filelist = {} # generate list of files to tweak -- by filter
	for fi in imgs:
		hdu = fits.open(fi)
		try: #get filter
			filt = hdu[0].header['FILTER']
		except:
			if hdu[0].header['FILTER1'].startswith('F'):
				filt = hdu[0].header['FILTER1']
			else:
				filt = hdu[0].header['FILTER2']
		if filt not in filelist.keys():
			filelist[filt] = []
		filelist[filt].append(fi)

	if not pretweaked:
		for fk in filelist.keys():
			filemodels = tweakreg_step.TweakRegStep(**tweakparams).call(filelist[fk], 
					output_dir = img_dir, save_results = True)

		imgs = sorted(glob.glob(img_dir+'*_tweakregstep.fits'))

	drizzlelist = {} #write file prefixes to drizzle per object per filter
	if genpsf: #generate model PSFs for each image + object
		if type(obj) == str: #check number of objects
			drizzlelist[obj] = {}
			skycoords = tools.objloc(obj)
			for i in imgs:
				pos = tools.checkpixloc(skycoords, i, inst, camera)

				coordstring = str(skycoords.ra)
				if skycoords.dec.deg >= 0:
					coordstring += '+'+str(skycoords.dec)
				if skycoords.dec.deg < 0:
					coordstring += str(skycoords.dec)

				modname = i.replace('.fits', '_'+coordstring+'_%s'%pos[3]+'_topsf.fits')
				if np.isfinite(pos[0]): #confirm object falls onto image
					if pos[3] not in drizzlelist[obj].keys():
						drizzlelist[obj][pos[3]] = []
					drizzlelist[obj][pos[3]].append(modname)

					psffunc(skycoords, i, imcam, pos, plot, verbose, **kwargs)

		if type(obj) != str: #if multiple objects, option to parallelize 
			skycoords = [] #only open each FITS file once

			for o in obj:
				drizzlelist[o] = {}
				skycoords.append(tools.objloc(o))
			
			for i in imgs:

				if parallel:
					if method.upper() == 'PSFEX':
						warnings.warn('Warning: Check your config and param files to ensure output files have unique names.', Warning, stacklevel = 2)
					pool = Pool(processes=(cpu_count() - 1))
					for j, coord in enumerate(skycoords):
						coordstring = str(sc.ra)
						if coord.dec.deg >= 0:
							coordstring += '+'+str(coord.dec)
						if coord.dec.deg < 0:
							coordstring += str(coord.dec)

						pos = tools.checkpixloc(coord, i, inst, camera)

						modname = i.replace('.fits', '_'+coordstring+'_%s'%pos[3]+'_topsf.fits')
						if np.isfinite(p[0]): #confirm that object falls onto detector
							if pos[3] not in drizzlelist[obj[j]].keys():
								drizzlelist[obj[j]][pos[3]] = []
							drizzlelist[obj[j]][pos[3]].append(modname)

							pool.apply_async(psffunc, args = (coord, i, imcam, pos, plot, verbose), kwds = kwargs)
					pool.close()
					pool.join()

				if not parallel:
					for j, coord in enumerate(skycoords):
						pos = tools.checkpixloc(coord, i, inst, camera)

						coordstring = str(coord.ra)
						if coord.dec.deg >= 0:
							coordstring += '+'+str(coord.dec)
						if coord.dec.deg < 0:
							coordstring += str(coord.dec)

						modname = i.replace('.fits', '_'+coordstring+'_%s'%pos[3]+'_topsf.fits')
						if np.isfinite(pos[0]): #confirm that object falls onto detector
							if pos[3] not in drizzlelist[obj[j]].keys():
								drizzlelist[obj[j]][pos[3]] = []
							drizzlelist[obj[j]][pos[3]].append(modname)

						psffunc(coord, i, imcam, pos, plot, verbose, **kwargs) 
					
	if not genpsf:
		userpsfs = sorted(glob.glob(usermethod))

		for up in userpsfs:
			im, obj, filt, _ = up.split('_')
			if obj not in drizzlelist.keys():
				drizzlelist[obj] = {}
			if filt not in drizzlelist[obj].keys():
				drizzlelist[obj][filt] = []
			drizzlelist[obj][filt].append(im)

	#####################################################################
	for do in drizzlelist.keys():
		if parallel:
			pool = Pool(processes=(cpu_count() - 1))
			for dk in drizzlelist[do].keys():

				shortdec, shortra = [cc.split('.')[0] for cc in do.split(' ')]

				if ':' not in shortra:
					if int(shortra) > 0:
						shortra = "+"+shortra

				resampname = shortdec+shortra+'_'+dk
				resampname = resampname.replace(':', '').replace(' ', '')

				resamp = resample_step.ResampleStep(**drizzleparams)
				resampkwds = {'input_models': drizzlelist[do][dk], 
							  'output_file': resampname,
							  'output_dir':img_dir, 
							  'save_results':True}
				pool.apply_async(resamp.call, kwds = resampkwds)
			pool.close()
			pool.join()
		if not parallel:
			for dk in drizzlelist[do].keys():
				shortdec, shortra = [cc.split('.')[0] for cc in do.split(' ')]

				if ':' not in shortra:
					if int(shortra) > 0:
						shortra = "+"+shortra

				resampname = shortdec+shortra+'_'+dk
				resampname = resampname.replace(':', '').replace(' ', '')

				resamp = resample_step.ResampleStep(**drizzleparams).call(drizzlelist[do][dk],
					output_file = resampname, 
					output_dir = img_dir, save_results = True)


	
	if drizzleimgs: # useful for processing all images + PSFs simultaneously
		for fk in filelist.keys():
			resamp = resample_step.ResampleStep(**drizzleparams).call(filelist[fk],
					output_file = '%s_img'%fk, output_dir = img_dir, save_results = True)

	#####################################################################
	suff = "resamplestep"

	if not finalonly:
		# clean up step to move all of the PSF files to the relevant directory
		# should grab all .pngs, .fits etc.
		if not os.path.exists(savedir):
			os.makedirs(savedir)
		os.system('mv %s*_%s* %s'%(img_dir, suff, savedir)) # generated PSF models
		os.system('mv %s*_psf %s'%(img_dir, savedir))
		os.system('mv %s*.psf %s'%(img_dir, savedir))
		os.system('mv %s*_topsf* %s'%(img_dir, savedir)) # tweaked and drizzled PSF models

		## clean up other files generated in the process
		os.system('mv %s*.cat %s'%(img_dir, savedir))
		os.system('mv %s*_mask.fits %s'%(img_dir, savedir))
		# retain tweaked version in working directory for re-runs etc.
		# os.system('mv %s*_tweakregstep.fits %s'%(img_dir, savedir))

		if verbose:
			print('Moved PSF files to %s'%savedir)

	if finalonly:
		# clean up step to move all of the resampled files to the relevant directory
		if not os.path.exists(savedir):
			os.makedirs(savedir)

		os.system('mv %s*_resamplestep* %s'%(img_dir, savedir)) #move files to preserve
		os.system('mv %s*_%s* %s'%(img_dir, suff, savedir)) # generic name for flexibility

		if verbose:
			print('Moved resampled files to %s'%savedir)


		if os.path.exists(removedir):
			# raise warning if removedir exists; not an error, though since some people might create it for this purpose
			warnings.warn('%s already exists. This directory and its contents will be deleted.'%(removedir),
				Warning, stacklevel = 2)

		if not os.path.exists(removedir): #directory to remove excess files
			os.makedirs(removedir)

		os.system('mv %s*_psf %s'%(img_dir, removedir))
		os.system('mv %s*.psf %s'%(img_dir, removedir))
		os.system('mv %s*_topsf* %s'%(img_dir, removedir))

		## clean up other files generated in the process
		os.system('mv %s*.cat %s'%(img_dir, removedir))
		os.system('mv %s*_mask.fits %s'%(img_dir, removedir))

		os.system('rm -r %s'%s(removedir))

		if verbose:
			print('Deleted intermediate products and removedir.')



	if out == 'asdf':
		# .asdf file read out in addition to .fits
		# defining suffix from resample output
		if savedir.split('/')[-1] != '':
			savedir += '/'

		dout = sorted(glob.glob(savedir+'*_%s.fits'%suff)) 
		for di in dout:
			tools.to_asdf(di)
		if verbose:
			print('Generated ASDF output')

	if returnpsf:
		returndict = {}
		for do in drizzlelist.keys():
			returndict[do] = {}
			for dk in drizzlelist[do].keys():

				shortdec, shortra = [cc.split('.')[0] for cc in do.split(' ')]
				if ':' not in shortra:
					if int(shortra) > 0:
						shortra = "+"+shortra

				resampname = shortdec+shortra+'_'+dk
				resampname = resampname.replace(':', '').replace(' ', '')

				if savedir.split('/')[-1] != '':
					savedir += '/'

				if returnpsf == 'full':	
					dr_psf = fits.open(savedir+'%s_%s.fits'%(resampname ,suff))
					returndict[do][dk] = dr_psf[1].data

				if returnpsf == 'crop':
					crop = tools.cutout(img = savedir+'%s_%s.fits'%(resampname ,suff), 
									coords = tools.objloc(do), fov_pixel = cutout_fov, save = savecutout)
					returndict[do][dk] = crop

		return returndict


def roman(img_dir, obj, inst, img_type= 'cal', file_type = 'fits', camera = None, method = 'WebbPSF', 
		usermethod = None, savedir = 'psfs', drizzleimgs = False, pretweaked = False, 
		usecrds = False, keeporig = True, plot = False, verbose = False, parallel = False, 
		out = 'fits', returnpsf = 'full', cutout_fov = 151, savecutout = True, finalonly = False,
		removedir = 'toremove', tweakparams = {}, drizzleparams = {}, **kwargs):
	"""
	Generate drizzled Roman Space Telescope PSFs.

	Parameters:
		img_dir (str): Path to directory containing calibrated files for which model PSF will be generated.
				If using the tweakreg step, best to include a drizzled file, as well, which can be used as a reference.
		obj(str, arr-like): Name or coordinates of object of interest in HH:MM:DD DD:MM:SS or degree format.
		img_type (str): e.g, 'cal' -- specifies which file-type to include.
		file_type (str): 'fits' or 'asdf' -- format to use for reading and manipulating data files.
		inst (str): 'WFI', 'CGI'
		camera (str): None
		method (str): 'WebbPSF', 'epsf' (empirical), 'PSFEx' (empirical) -- see spike.psfgen for details -- or 'USER';
				if 'USER', usermethod should be a function that generates, or path to a directory of user-generated, PSFs 
				named [imgprefix]_[coords]_[band]_psf.fits, e.g., imgprefix_23.31+30.12_F814W_psf.fits or 
				imgprefix_195.78-46.52_F555W_psf.fits. Note: the WebbPSF name is maintained here in lieu of STPSF to avoid
				confusion with the generation of empirical STDPSFs.
		usermethod (func or str): If method = 'USER', usermethod should be a function that generates, or path to a 
				directory of user-generated, PSFs named [imgprefix]_[coords]_[band]_psf.fits, e.g., 
				imgprefix_23.31+30.12_F814W_psf.fits or imgprefix_195.78-46.52_F555W_psf.fits, where the 
				imgprefix corresponds to the name of the relevant cal/calints... files in the directory and the 
				headers are from the original images (see spike.tools.rewrite_fits, which can be used to this end).
		savedir (str): Where the PSF models and drizzled PSF will be saved. Defaults to 'psfs'.
		drizzleimgs (bool): If True, will drizzle the input images at the same time as creating a drizzled psf.
		pretweaked (bool): If True, skips tweak step to include fine WCS corrections.
		usecrds (bool): If True, use CRDS config settings as defaults. 
		keeporig (bool): If True (and pretweaked = False), create copy of img_dir before tweak.
		plot (bool): If True, saves .pngs of the model PSFs.
		verbose (bool): If True, prints progress messages.
		parallel (bool): If True, runs PSF generation in parallel.
		out (str): 'fits' or 'asdf'. Output for the drizzled PSF. If 'asdf', .asdf AND .fits are saved.
		returnpsf (str): 'full', 'crop', or None. If None, spike.psf.roman does not return anything. If 'full' (default),
			returns the PSF in the full spatial context of the processed image. If 'crop', returns the region immediately
			around the PSF (size of cutout set by cutout_fov).
		cutout_fov (int): Side length in pixels of square cutout region centered on PSF. Used if returnpsf = 'crop'.
		savecutout (bool): If True, save a .fits file with the cutout region, including WCS.
		finalonly (bool): If True, only retains final drizzled/resampled data products in savedir and deletes intermediate products.
		removedir (str): Directory (**to be deleted**) that stores intermediate products for removal. Default is 'toremove'.
		tweakparams (dict): Dictionary of keyword arguments for drizzlepac.tweakreg. See the drizzlepac documentation
				for a full list.
		drizzleparams (dict): Dictionary of keyword arguments for drizzlepac.astrodrizzle. See the drizzlepac 
				documentation for a full list.
		**kwargs: Keyword arguments for PSF generation function.

	Returns:
		Generates model PSFs and drizzled PSF. (If drizzledimgs = True, also produces drizzled image from input files.)

		If returnpsf = 'full', will return each of the full drizzled PSF images in an object, filter indexed dict.
		If returnpsf = 'crop', will return a cutout region of the drizzled PSF images (around the PSF) in an obj, filt indexed dict.
	"""
	os.environ['CRDS_SERVER_URL']="https://roman-crds.stsci.edu"

	if not usecrds:
		os.environ["STPIPE_DISABLE_CRDS_STEPPARS"] = 'True'

	from spike.romancal import tweakreg_step, resample_step

	if img_dir[:-1] != '/':
		img_dir += '/' #force paths to work out

	if keeporig and not pretweaked:
		if not os.path.exists(img_dir+'_orig'):
			os.makedirs(img_dir+'_orig')
		os.system('cp -r '+img_dir+'*_'+img_type+'.fits '+'img_dir'+'_orig')
		if verbose:
			print('Made copy of '+img_dir)


	imgs = sorted(glob.glob(img_dir+'*'+img_type+'.fits'))

	imcam = inst.upper()

	genpsf = True
	if method.upper() not in ['WEBBPSF', 'EPSF', 'PSFEX', 'USER']:
		raise Exception('tool must be one of WEBBPSF, EPSF, PSFEX, USER')
	if method.upper() == 'WEBBPSF':
		psffunc = psfgen.jwpsf
	if method.upper() == 'EPSF':
		psffunc = psfgen.effpsf
	if method.upper() == 'PSFEX':
		psffunc = psfgen.psfex
	if method.upper() == 'USER':
		if type(usermethod) == str: #check if user input is path to directory
			genpsf = False
		if type(usermethod) != str: #or function
			psffunc = method

	filelist = {} # generate list of files to tweak -- by filter
	for fi in imgs:
		hdu = fits.open(fi)
		try: #get filter
			filt = hdu[0].header['FILTER']
		except:
			if hdu[0].header['FILTER1'].startswith('F'):
				filt = hdu[0].header['FILTER1']
			else:
				filt = hdu[0].header['FILTER2']
		if filt not in filelist.keys():
			filelist[filt] = []
		filelist[filt].append(fi)

	if not pretweaked:
		for fk in filelist.keys():
			filemodels = tweakreg_step.TweakRegStep(**tweakparams).call(filelist[fk], 
					output_dir = img_dir, save_results = True)

		imgs = sorted(glob.glob(img_dir+'*_tweakregstep.fits'))

	drizzlelist = {} #write file prefixes to drizzle per object per filter
	if genpsf: #generate model PSFs for each image + object
		if type(obj) == str: #check number of objects
			drizzlelist[obj] = {}
			skycoords = tools.objloc(obj)
			for i in imgs:
				pos = tools.checkpixloc(skycoords, i, inst, camera)
				coordstring = str(skycoords.ra)
				if skycoords.dec.deg >= 0:
					coordstring += '+'+str(skycoords.dec)
				if skycoords.dec.deg < 0:
					coordstring += str(skycoords.dec)

				modname = i.replace('.fits', '_'+coordstring+'_%s'%pos[3]+'_topsf.fits')
				if np.isfinite(pos[0]): #confirm object falls onto image
					if pos[3] not in drizzlelist[obj].keys():
						drizzlelist[obj][pos[3]] = []
					drizzlelist[obj][pos[3]].append(modname)

					psffunc(skycoords, i, imcam, pos, plot, verbose, **kwargs)

		if type(obj) != str: #if multiple objects, option to parallelize 
			skycoords = [] #only open each FITS file once

			for o in obj:
				drizzlelist[o] = {}
				skycoords.append(tools.objloc(o))
			
			for i in imgs:

				pos = tools.checkpixloc(skycoords, i, inst, camera)

				if parallel:
					if method.upper() == 'PSFEX':
						warnings.warn('Warning: Check your config and param files to ensure output files have unique names.', Warning, stacklevel = 2)
					pool = Pool(processes=(cpu_count() - 1))
					for j, p in enumerate(pos):
						coordstring = str(skycoords[j].ra)
						if skycoords[j].dec.deg >= 0:
							coordstring += '+'+str(skycoords[j].dec)
						if skycoords[j].dec.deg < 0:
							coordstring += str(skycoords[j].dec)
							
						modname = i.replace('.fits', '_'+coordstring+'_%s'%p[3]+'_topsf.fits')
						if np.isfinite(p[0]): #confirm that object falls onto detector
							if p[3] not in drizzlelist[obj[j]].keys():
								drizzlelist[obj[j]][p[3]] = []
							drizzlelist[obj[j]][p[3]].append(i)

							pool.apply_async(psffunc, args = (skycoords[j], i, imcam, p, plot, verbose), kwds = kwargs)
					pool.close()
					pool.join()
				if not parallel:
					for j, coord in enumerate(skycoords):
						psffunc(coord, i, imcam, pos[j], plot, verbose, **kwargs) 
					
	if not genpsf:
		userpsfs = sorted(glob.glob(usermethod))

		for up in userpsfs:
			im, obj, filt, _ = up.split('_')
			if obj not in drizzlelist.keys():
				drizzlelist[obj] = {}
			if filt not in drizzlelist[obj].keys():
				drizzlelist[obj][filt] = []
			drizzlelist[obj][filt].append(im)

	#####################################################################
	for do in drizzlelist.keys():
		if parallel:
			pool = Pool(processes=(cpu_count() - 1))
			for dk in drizzlelist[do].keys():
				shortdec, shortra = [cc.split('.')[0] for cc in do.split(' ')]

				if ':' not in shortra:
					if int(shortra) > 0:
						shortra = "+"+shortra

				resampname = shortdec+shortra+'_'+dk
				resampname = resampname.replace(':', '').replace(' ', '')

				resamp = resample_step.ResampleStep(**drizzleparams)
				resampkwds = {'input_models': drizzlelist[do][dk], 
							  'output_file': resampname,
							  'output_dir':img_dir, 
							  'save_results':True}
				pool.apply_async(resamp.call, kwds = resampkwds)
			pool.close()
			pool.join()
		if not parallel:
			for dk in drizzlelist[do].keys():
				shortdec, shortra = [cc.split('.')[0] for cc in do.split(' ')]

				if ':' not in shortra:
					if int(shortra) > 0:
						shortra = "+"+shortra

				resampname = shortdec+shortra+'_'+dk
				resampname = resampname.replace(':', '').replace(' ', '')

				resamp = resample_step.ResampleStep(**drizzleparams).call(drizzlelist[do][dk],
					output_file = resampname, output_dir = img_dir, save_results = True)


	
	if drizzleimgs: # useful for processing all images + PSFs simultaneously
		for fk in filelist.keys():
			resamp = resample_step.ResampleStep(**drizzleparams).call(filelist[fk],
					output_file = '%s_img'%fk, output_dir = img_dir, save_results = True)
	#####################################################################
	suff = "resamplestep"

	if not finalonly:
		# clean up step to move all of the PSF files to the relevant directory
		# should grab all .pngs, .fits etc.
		if not os.path.exists(savedir):
			os.makedirs(savedir)
		os.system('mv %s*_%s* %s'%(img_dir, suff, savedir)) # generated PSF models
		os.system('mv %s*_psf %s'%(img_dir, savedir))
		os.system('mv %s*.psf %s'%(img_dir, savedir))
		os.system('mv %s*_topsf* %s'%(img_dir, savedir)) # tweaked and drizzled PSF models

		## clean up other files generated in the process
		os.system('mv %s*.cat %s'%(img_dir, savedir))
		os.system('mv %s*_mask.fits %s'%(img_dir, savedir))
		# retain tweaked version in working directory for re-runs etc.
		# os.system('mv %s*_tweakregstep.fits %s'%(img_dir, savedir))

		if verbose:
			print('Moved PSF files to %s'%savedir)

	if finalonly:
		# clean up step to move all of the resampled files to the relevant directory
		if not os.path.exists(savedir):
			os.makedirs(savedir)

		os.system('mv %s*_resamplestep* %s'%(img_dir, savedir)) #move files to preserve
		os.system('mv %s*_%s* %s'%(img_dir, suff, savedir)) # generic name for flexibility

		if verbose:
			print('Moved resampled files to %s'%savedir)


		if os.path.exists(removedir):
			# raise warning if removedir exists; not an error, though since some people might create it for this purpose
			warnings.warn('%s already exists. This directory and its contents will be deleted.'%(removedir),
				Warning, stacklevel = 2)

		if not os.path.exists(removedir): #directory to remove excess files
			os.makedirs(removedir)

		os.system('mv %s*_psf %s'%(img_dir, removedir))
		os.system('mv %s*.psf %s'%(img_dir, removedir))
		os.system('mv %s*_topsf* %s'%(img_dir, removedir))

		## clean up other files generated in the process
		os.system('mv %s*.cat %s'%(img_dir, removedir))
		os.system('mv %s*_mask.fits %s'%(img_dir, removedir))

		os.system('rm -r %s'%s(removedir))

		if verbose:
			print('Deleted intermediate products and removedir.')


	if out == 'asdf':
		# .asdf file read out in addition to .fits
		# defining suffix from resample output
		if savedir.split('/')[-1] != '':
			savedir += '/'

		dout = sorted(glob.glob(savedir+'*_%s.fits'%s)) 
		for di in dout:
			tools.to_asdf(di)


	if returnpsf:
		returndict = {}
		for do in drizzlelist.keys():
			returndict[do] = {}
			for dk in drizzlelist[do].keys():

				shortdec, shortra = [cc.split('.')[0] for cc in do.split(' ')]
				if ':' not in shortra:
					if int(shortra) > 0:
						shortra = "+"+shortra

				resampname = shortdec+shortra+'_'+dk
				resampname = resampname.replace(':', '').replace(' ', '')

				if savedir.split('/')[-1] != '':
					savedir += '/'

				if returnpsf == 'full':	
					dr_psf = fits.open(savedir+'%s_%s.fits'%(resampname, suff))
					returndict[do][dk] = dr_psf[1].data

				if returnpsf == 'crop':
					crop = tools.cutout(img = savedir+'%s_%s.fits'%(resampname, suff), 
									coords = tools.objloc(do), fov_pixel = cutout_fov, save = savecutout)
					returndict[do][dk] = crop

		return returndict




"""
This module is designed to make stpipe functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


stpipe copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/stpipe/blob/main/src/stpipe/log.py
"""

"""
Logging setup etc.
"""

import fnmatch
import io
import logging
import os
import pathlib
import sys
import threading
from contextlib import contextmanager

from astropy.extern.configobj import validate
from astropy.extern.configobj.configobj import ConfigObj

from spike.stpipe import config_parser

STPIPE_ROOT_LOGGER = "stpipe"
DEFAULT_FORMAT = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
DEFAULT_CONFIGURATION = b"""
[*]
handler = stderr
level = INFO
"""

MAX_CONFIGURATION = b"""
[*]
handler = stderr
level = DEBUG
"""


###########################################################################
# LOGS AS EXCEPTIONS


class LoggedException(Exception):  # noqa: N818
    """
    This is an exception used when a log record is converted into an
    exception.

    Use its `record` member to get the original `logging.LogRecord`.
    """

    def __init__(self, record):
        self.record = record
        Exception.__init__(self, record.getMessage())


class BreakHandler(logging.Handler):
    """
    A handler that turns logs of a certain severity or higher into
    exceptions.
    """

    _from_config = True

    def emit(self, record):
        raise LoggedException(record)


#########################################################################
# LOGGING CONFIGURATION

# A dictionary mapping patterns to
log_config = {}


class LogConfig:
    """
    Stores a single logging configuration.

    Parameters
    ----------
    name : str
        The `fnmatch` pattern used to match the logging class

    handler, level, break_level, format : str
        See LogConfig.spec for a description of these values.
    """

    def __init__(
        self,
        name,
        handler=None,
        level=logging.NOTSET,
        break_level=logging.NOTSET,
        format=None,  # noqa: A002
    ):
        if name in ("", ".", "root"):
            name = "*"
        self.name = name
        self.handler = handler
        if not isinstance(self.handler, list):
            if self.handler.strip() == "":
                self.handler = []
            else:
                self.handler = [x.strip() for x in self.handler.split(",")]
        self.level = level
        self.break_level = break_level
        if format is None:
            format = DEFAULT_FORMAT  # noqa: A001
        self.format = format

    def match(self, log_name):
        """
        Returns `True` if `log_name` matches the pattern of this
        configuration.
        """
        if log_name.startswith(STPIPE_ROOT_LOGGER):
            log_name = log_name[len(STPIPE_ROOT_LOGGER) + 1 :]
            if fnmatch.fnmatchcase(log_name, self.name):
                return True
        return False

    def get_handler(self, handler_str):
        """
        Given a handler string, returns a `logging.Handler` object.
        """
        if handler_str.startswith("file:"):
            return logging.FileHandler(handler_str[5:], "w", "utf-8", True)

        if handler_str.startswith("append:"):
            return logging.FileHandler(handler_str[7:], "a", "utf-8", True)

        if handler_str == "stdout":
            return logging.StreamHandler(sys.stdout)

        if handler_str == "stderr":
            return logging.StreamHandler(sys.stderr)

        raise ValueError(f"Can't parse handler {handler_str!r}")

    def apply(self, log):
        """
        Applies the configuration to the given `logging.Logger`
        object.
        """
        for handler in log.handlers[:]:
            if hasattr(handler, "_from_config"):
                log.handlers.remove(handler)

        # Set a handler
        for handler_str in self.handler:
            handler = self.get_handler(handler_str)
            handler._from_config = True
            handler.setLevel(self.level)
            log.addHandler(handler)

        # Set the log level
        log.setLevel(self.level)

        # Set the break level
        if self.break_level != logging.NOTSET:
            log.addHandler(BreakHandler(self.break_level))

        formatter = logging.Formatter(self.format)
        for handler in log.handlers:
            if isinstance(handler, logging.Handler) and hasattr(
                handler, "_from_config"
            ):
                handler.setFormatter(formatter)

    def match_and_apply(self, log):
        """
        If the given `logging.Logger` object matches the pattern of
        this configuration, it applies the configuration to it.
        """
        if self.match(log.name):
            self.apply(log)


def load_configuration(config_file):
    """
    Loads a logging configuration file.  The format of this file is
    defined in LogConfig.spec.

    Parameters
    ----------
    config_file : str, pathlib.Path instance or readable file-like object
    """

    def _level_check(value):
        try:
            value = int(value)
        except ValueError:
            pass

        try:
            value = logging._checkLevel(value)
        except ValueError as err:
            raise validate.VdtTypeError(value) from err
        return value

    spec = config_parser.load_spec_file(LogConfig)
    if isinstance(config_file, pathlib.Path):
        config_file = str(config_file)
    config = ConfigObj(config_file, raise_errors=True, interpolation=False)
    val = validate.Validator()
    val.functions["level"] = _level_check
    config_parser.validate(config, spec, validator=val)

    log_config.clear()

    for key, val in config.items():
        log_config[key] = LogConfig(key, **val)

    for log in logging.Logger.manager.loggerDict.values():
        if isinstance(log, logging.Logger):
            for cfg in log_config.values():
                cfg.match_and_apply(log)


def getLogger(name=None):  # noqa: N802
    return logging.getLogger(name)


def _find_logging_config_file():
    files = ["stpipe-log.cfg", "~/.stpipe-log.cfg", "/etc/stpipe-log.cfg"]

    for file in files:
        file = os.path.expanduser(file)
        if os.path.exists(file):
            return os.path.abspath(file)

    return io.BytesIO(DEFAULT_CONFIGURATION)


###########################################################################
# LOGGING DELEGATION


class DelegationHandler(logging.Handler):
    """
    A handler that delegates messages along to the currently active
    `Step` logger.  It only delegates messages that come from outside
    of the `stpipe` hierarchy, in order to prevent infinite recursion.

    Since we could be multi-threaded and each thread may be running a
    different thread, we need to manage a dictionary mapping the
    current thread to the Step's logger on that thread.
    """

    def __init__(self, *args, **kwargs):
        self._logs = {}
        logging.Handler.__init__(self, *args, **kwargs)

    def emit(self, record):
        log = self.log
        if log is not None and not record.name.startswith(STPIPE_ROOT_LOGGER):
            record.name = log.name
            log.handle(record)

    @property
    def log(self):
        return self._logs.get(threading.current_thread(), None)

    @log.setter
    def log(self, log):
        if log is not None and not (
            isinstance(log, logging.Logger) and log.name.startswith(STPIPE_ROOT_LOGGER)
        ):
            raise AssertionError("Can't set the log to a root logger")

        self._logs[threading.current_thread()] = log


class RecordingHandler(logging.Handler):
    """
    A handler that simply accumulates LogRecord instances.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._log_records = []

    @property
    def log_records(self):
        return self._log_records

    def emit(self, record):
        if self.formatter is not None:
            self._log_records.append(self.formatter.format(record))


@contextmanager
def record_logs(level=logging.NOTSET, formatter=None):
    if formatter is None:
        yield []
    else:
        handler = RecordingHandler(level=level)
        handler.setFormatter(formatter)
        logger = getLogger(STPIPE_ROOT_LOGGER)
        logger.addHandler(handler)
        try:
            yield handler.log_records
        finally:
            logger.removeHandler(handler)


# Install the delegation handler on the root logger.  The Step class
# uses the `delegator` instance to change what the current Step logger
# is.
log = getLogger()
delegator = DelegationHandler()
delegator.log = getLogger(STPIPE_ROOT_LOGGER)
log.addHandler(delegator)

logging_config_file = _find_logging_config_file()
if logging_config_file is not None:
    load_configuration(logging_config_file)

logging.captureWarnings(True)




"""
This module is designed to make stpipe functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


stpipe copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/jwst/blob/main/jwst/datamodels/library.py
"""

import abc
import copy
import json
import os.path
import tempfile
import warnings
from collections.abc import Iterable, MutableMapping
from pathlib import Path
from types import MappingProxyType

import asdf

from spike.stpipe.datamodel import AbstractDataModel

__all__ = [
    "LibraryError",
    "BorrowError",
    "ClosedLibraryError",
    "NoGroupID",
    "AbstractModelLibrary",
]


class LibraryError(Exception):
    """
    Generic ModelLibrary related exception
    """


class BorrowError(LibraryError):
    """
    Exception indicating an issue with model borrowing
    """


class ClosedLibraryError(LibraryError):
    """
    Exception indicating a library method was used outside of a
    ``with`` context (that "opens" the library).
    """


class NoGroupID(LibraryError):
    """
    Exception to use when a model has no "group_id".
    """


class _Ledger(MutableMapping):
    """
    A "ledger" used for tracking borrowed out models.

    Each model has a unique "index" in the library because
    the order of the models in the library never changes.
    This "index" can be used to track the model.

    For ease-of-use this ledger maintains 2 mappings:

        - id (the id(model) result) to model index
        - index to model

    The "index to model" mapping keeps a reference to every
    model in the ledger (which allows id(model) to be consistent).

    The ledger is a MutableMapping that supports look up of:
        - index for a model
        - model for an index
    """

    def __init__(self):
        self._id_to_index = {}
        self._index_to_model = {}

    def __getitem__(self, model_or_index):
        if isinstance(model_or_index, int):
            return self._index_to_model[model_or_index]
        return self._id_to_index[id(model_or_index)]

    def __setitem__(self, index, model):
        self._index_to_model[index] = model
        self._id_to_index[id(model)] = index

    def __delitem__(self, model_or_index):
        if isinstance(model_or_index, int):
            index = model_or_index
            model = self._index_to_model[index]
        else:
            model = model_or_index
            index = self._id_to_index[id(model)]
        del self._id_to_index[id(model)]
        del self._index_to_model[index]

    def __iter__(self):
        # only return indexes
        return iter(self._index_to_model)

    def __len__(self):
        return len(self._id_to_index)


class AbstractModelLibrary(abc.ABC):
    """
    A "library" of models (loaded from an association file).

    Do not anger the librarian!

    The library owns all models from the association and it will handle
    opening and closing files.

    Models can be "borrowed" from the library (by iterating through the
    library or "borrowing" a specific model). However the library must be
    "open" (used in a ``with`` context)  to borrow a model and the model
    must be "shelved" before the library "closes" (the ``with`` context exits).

    >>> with library:   # doctest: +SKIP
            model = library.borrow(0)  # borrow the first model
            # do stuff with the model
            library.shelve(model, 0)  # return the model

    Failing to "open" the library will result in a `ClosedLibraryError`.

    Failing to "return" a borrowed model will result in a `BorrowError`.
    """

    def __init__(
        self,
        init,
        asn_exptypes=None,
        asn_n_members=None,
        on_disk=False,
        temp_directory=None,
        **datamodels_open_kwargs,
    ):
        """
        Create a new ModelLibrary based on the provided "init".

        Parameters
        ----------
        init : str or Path or list (of DataModels)
            If a string or Path this should point to a valid
            association file. If a list of models consider continuing
            to use the list instead of making a library (the list will
            be more efficient and easier to use).

        asn_exptypes : list of str, optional
            List of "exptypes" to load from the "init" value. Any
            association member with a "exptype" that matches (case
            insensitive) a value in this list will be available
            through the library.

        asn_n_members : int, optional
            Number of association members to include in the library.
            This filtering will occur after "asn_exptypes" is applied.
            By default all members will be included.

        on_disk : bool, optional, default=False
            When enabled, keep the models in the library "on disk",
            writing out temporary files when the models are "shelved".
            This option is only compatible with models loaded from
            an association (not a list of models).

        temp_directory : str or Path, optional
            The temporary directory in which to store models when
            "on_disk" is enabled. By default a
            ``tempfile.TemporaryDirectory`` will be created in the
            working directory.

        **datamodels_open_kwargs : dict
            Keyword arguments that will provided to ``_datamodels_open``
            when models are opened.
        """
        self._on_disk = on_disk
        self._open = False
        self._ledger = _Ledger()
        self._loaded_models = {}

        self._datamodels_open_kwargs = datamodels_open_kwargs

        if on_disk:
            if temp_directory is None:
                self._temp_dir = tempfile.TemporaryDirectory(dir="")
                self._temp_path = Path(self._temp_dir.name)
            else:
                self._temp_path = Path(temp_directory)
            self._temp_filenames = {}

        if asn_exptypes is not None:
            # if a single string, treat this as a single item list
            if isinstance(asn_exptypes, str):
                asn_exptypes = [asn_exptypes]
            # convert these to lower case to allow case insensitive matching below
            asn_exptypes = [exptype.lower() for exptype in asn_exptypes]

        if isinstance(init, (str, Path)):
            # init is an association filename (or path)
            asn_path = os.path.abspath(os.path.expanduser(os.path.expandvars(init)))
            self._asn_dir = os.path.dirname(asn_path)

            # load association
            asn_data = self._load_asn(asn_path)

            # store the basename of the input file as the table_name
            asn_data["table_name"] = os.path.basename(asn_path)
        elif isinstance(init, MutableMapping):
            # init is an association "dictionary"
            # we will modify the asn below so do a deep copy
            asn_data = copy.deepcopy(init)
            self._asn_dir = os.path.abspath(".")
        elif isinstance(init, self.__class__):
            raise ValueError(f"Invalid init {init}")
        elif isinstance(init, Iterable):  # assume a list of models
            if on_disk:
                raise ValueError("on_disk cannot be used for a list of models")

            # init is a list of models, this type of input ignores much of
            # the benefits of this class but is common in test code.

            # make a fake asn from the models
            members = []
            for index, model_or_filename in enumerate(init):
                if asn_n_members is not None and len(members) == asn_n_members:
                    break
                if isinstance(model_or_filename, (str, Path)):
                    # Supporting a list of filenames by opening them as models
                    # has issues, if this is a widely supported mode (vs providing
                    # an association) it might make the most sense to make a fake
                    # association with the filenames at load time.
                    model = self._datamodels_open(
                        model_or_filename, **self._datamodels_open_kwargs
                    )
                else:
                    model = model_or_filename

                exptype = self._model_to_exptype(model)

                if asn_exptypes is not None and exptype.lower() not in asn_exptypes:
                    continue

                members.append(
                    {
                        "expname": self._model_to_filename(model),
                        "exptype": exptype,
                        "group_id": self._to_group_id(model, len(members)),
                    }
                )

                self._loaded_models[len(self._loaded_models)] = model

            # since we've already filtered by asn type and n members reset these values
            asn_exptypes = None
            asn_n_members = None

            # make a very limited fake association
            asn_data = {
                "products": [
                    {
                        "members": members,
                    }
                ],
            }
            self._asn_dir = None
        else:
            raise ValueError(f"Invalid init {init}")

        if asn_exptypes is not None:
            asn_data["products"][0]["members"] = [
                m
                for m in asn_data["products"][0]["members"]
                if m["exptype"] in asn_exptypes
            ]

        if asn_n_members is not None:
            asn_data["products"][0]["members"] = asn_data["products"][0]["members"][
                :asn_n_members
            ]

        self._asn = asn_data
        self._members = self._asn["products"][0]["members"]

        for index, member in enumerate(self._members):
            if "group_id" not in member:
                filename = os.path.join(self._asn_dir, member["expname"])
                member["group_id"] = self._to_group_id(filename, index)

        if not on_disk:
            # if models were provided as input, assign the members here
            # now that the fake asn data is complete
            for i in self._loaded_models:
                self._assign_member_to_model(self._loaded_models[i], self._members[i])

    def __del__(self):
        if hasattr(self, "_temp_dir"):
            self._temp_dir.cleanup()

    @property
    def asn(self):
        """
        Association dictionary used to create the library.

        Note that changes to the models may cause this information
        to fall "out-of-sync" with the models. For example, borrowing
        and changing a models "group_id" metadata will not change
        the "group_id" entry for the member at the same index in the
        association dictionary returned for this property.

        This is "read-only" to reduce the chances that the
        information in this dictionary will conflict with
        the model metadata.
        """

        # return a "read only" association
        def _to_read_only(obj):
            if isinstance(obj, dict):
                return MappingProxyType(obj)
            if isinstance(obj, list):
                return tuple(obj)
            return obj

        return asdf.treeutil.walk_and_modify(self._asn, _to_read_only)

    @property
    def group_names(self):
        """
        A ``set`` of association member "group_id" values

        This property (similar to ``asn``) can fall "out-of-sync"
        with the model metadata. The names returned here will
        only reflect the "group_id" at the time of library creation.
        """
        names = set()
        for member in self._members:
            names.add(member["group_id"])
        return names

    @property
    def group_indices(self):
        """
        A ``dict`` with "group_id" as keys. Each value contains
        a ``list`` of indices of members that shared that "group_id".

        Note that this is based on the member entries in the association
        used to create the ModelLibrary. Updating the "group_id" of a model
        in the library will NOT change the values returned by this property
        (and will not change the association data available at ``asn``).
        """
        group_dict = {}
        for i, member in enumerate(self._members):
            group_id = member["group_id"]
            if group_id not in group_dict:
                group_dict[group_id] = []
            group_dict[group_id].append(i)
        return group_dict

    def __len__(self):
        return len(self._members)

    def borrow(self, index):
        """
        "borrow" a model from the library.

        Every model "borrowed" from the library must be returned before
        the library is closed (see the exceptions described below).

        For an "on_disk" library a calling this function will cause
        the corresponding model to be loaded from disk.

        Parameters
        ----------
        index : int
            The index of the model within the library. As the library
            size does not change this is unique to this model for this
            library.

        Returns
        -------
        model : DataModel

        Raises
        ------
        `ClosedLibraryError`
            If the library is not "open" (used in a ``with`` context).

        `BorrowError`
            If the model at this index is already "borrowed".
        """
        if not self._open:
            raise ClosedLibraryError("ModelLibrary is not open")

        # if model was already borrowed, raise
        if index in self._ledger:
            raise BorrowError("Attempt to double-borrow model")

        if self._on_disk:
            if index in self._temp_filenames:
                model = self._datamodels_open(
                    self._temp_filenames[index], **self._datamodels_open_kwargs
                )
            else:
                model = self._load_member(index)
        else:
            if index in self._loaded_models:
                # if this model is in memory, return it
                model = self._loaded_models[index]
            else:
                model = self._load_member(index)
                self._loaded_models[index] = model

        self._ledger[index] = model
        return model

    def _temp_path_for_model(self, model, index):
        """
        Determine the temporary path to use to save a model
        at the provided index in the library.
        """
        model_filename = self._model_to_filename(model)
        subpath = self._temp_path / f"{index}"
        if not os.path.exists(subpath):
            os.makedirs(subpath)
        return subpath / model_filename

    def shelve(self, model, index=None, modify=True):
        """
        "shelve" a model, returning it to the library.

        All borrowed models must be "shelved" before the library is
        closed (see notes about exceptions below).

        For an "on_disk" model shelving a model may cause
        a temporary file to be written (see "modify" below).

        Parameters
        ----------
        model : DataModel
            DataModel to return to the library at the provided index.

        index : int, optional
            The index within the library where the model will be stored.
            If the same model (checked by ``id(model)``) was previously
            borrowed and the index is not provided, the borrowed index
            will be used. If providing a new DataModel (one not borrowed
            from the library) the index must be provided, and the new
            model replaces the prior model at that index.

        modify : bool, optional, default=True
            For an "on_disk" library, temporary files will only be written
            when modify is True. For a library that is not "on_disk"
            this option has no effect and any modifications made to the model
            while it was borrowed will persist.

        Raises
        ------
        `ClosedLibraryError`
            If the library is not "open" (used in a ``with`` context).

        `BorrowError`
            If an unknown model is provided (without an index) or
            if the model at the provided index has not been "borrowed"
        """
        if not self._open:
            raise ClosedLibraryError("ModelLibrary is not open")

        if index is None:
            try:
                index = self._ledger[model]
            except KeyError:
                raise BorrowError("Attempt to shelve an unknown model")

        if index not in self._ledger:
            raise BorrowError("Attempt to shelve model at a non-borrowed index")

        if modify:
            if self._on_disk:
                temp_filename = self._temp_path_for_model(model, index)
                model.save(temp_filename)

                # if we have an old model for this index that was saved
                # in the temporary directory and this model has a different
                # filename, remove the old file.
                if index in self._temp_filenames:
                    old_filename = self._temp_filenames[index]
                    if old_filename != temp_filename:
                        os.remove(old_filename)

                self._temp_filenames[index] = temp_filename
            else:
                self._loaded_models[index] = model

        del self._ledger[index]

    def __iter__(self):
        """
        Iterate through the models in an (open) library by "borrowing"
        each one, one at a time.

        Returns
        -------
        model_iter : generator
            Generator that returns "borrowed" models. Failing to "shelve"
            the models produced by this generator will result in a
            `BorrowError`.
        """
        for i in range(len(self)):
            yield self.borrow(i)

    def _assign_member_to_model(self, model, member):
        """
        Assign association member information to an opened model.

        This will be called:

            - when a model is loaded (for the first time) in ``_load_member``
            - when the library is created (if the library was provided a
              list of open models)

        This will assign the following metadata attributes (setting
        for example ``model.meta.group_id`` to the value of
        ``member["group_id"]``):

            - group_id
            - exptype
            - tweakreg_catalog

        and (when available from the association) assign the following
        metadata attributes based on the association information:

            - set ``meta.asn.table_name`` to ``asn["table_name"]``
            - set ``meta.asn.pool_name`` to ``asn["pool_name"]``

        Parameters
        ----------
        model : DataModel
            Model to be updated (in place).

        member : dict
            Dictionary containing contents of the "member" entry
            (include "exptype", "expname", "group_id", etc...)
        """
        for attr in ("group_id", "tweakreg_catalog", "exptype"):
            if attr in member:
                setattr(model.meta, attr, member[attr])

        if not hasattr(model.meta, "asn"):
            model.meta.asn = {}

        if "table_name" in self.asn:
            model.meta.asn["table_name"] = self.asn["table_name"]
        if "asn_pool" in self.asn:
            model.meta.asn["pool_name"] = self.asn["asn_pool"]

    def _load_member(self, index):
        """
        Load a model for the association member at the provided index.

        This will be called when the model at the provided index:
            - is not already loaded
            - does not have a temporary file

        Parameters
        ----------
        index : int

        Returns
        -------
        model : DataModel
        """
        member = self._members[index]
        filename = os.path.join(self._asn_dir, member["expname"])

        model = self._datamodels_open(filename, **self._datamodels_open_kwargs)

        self._assign_member_to_model(model, member)

        return model

    def _save(self, path, **kwargs):
        """
        .. warning:: This save is NOT used by Step/Pipeline. This is
                     intentional as the Step/Pipeline has special requirements.

        For now this is a very basic "save". It does not:

            - check that the library contains no duplicate filenames
            - propagate asn_pool and other non-member asn information

        It will save:
            - models with filenames determined by ``_model_to_filename``
            - a very simple association that contains only 1 product that
              lists ``expname`` ``exptype`` and ``group_id``  for each model
              in a file named "asn.json".

        Parameters
        ----------
        path : Path or str
            Directory in which to store the models and generated
            association for this library.

        Returns
        -------
        association_path : Path
            The path to the saved association file.
        """
        if isinstance(path, str):
            path = Path(path)
        if not path.exists():
            path.mkdir()
        members = []
        with self:
            for i, model in enumerate(self):
                mfn = Path(self._model_to_filename(model))
                model.save(path / mfn, **kwargs)
                members.append(
                    {
                        "expname": str(mfn),
                        "exptype": self._model_to_exptype(model),
                        "group_id": self._model_to_group_id(model),
                    }
                )
                self.shelve(model, i, modify=False)
        asn_data = {"products": [{"members": members}]}
        asn_path = path / "asn.json"
        with open(asn_path, "w") as f:
            json.dump(asn_data, f)
        return asn_path

    def get_crds_parameters(self):
        """
        Get the "crds_parameters" from either:
            - the first "science" member (based on exptype)
            - the first model (if no "science" member is found)

        If no "science" members are found in the library a ``UserWarning``
        will be issued.

        Returns
        -------
        crds_parameters : dict
            The result of ``get_crds_parameters`` called on the selected
            model.
        """
        with self:
            science_index = None
            for i, member in enumerate(self._members):
                if member["exptype"].lower() == "science":
                    science_index = i
                    break
            if science_index is None:
                warnings.warn(
                    "get_crds_parameters failed to find any science members. "
                    "The first model was used to determine the parameters",
                    UserWarning,
                )
                science_index = 0
            model = self.borrow(science_index)
            parameters = model.get_crds_parameters()
            self.shelve(model, science_index, modify=False)
        return parameters

    def finalize_result(self, step, reference_files_used):
        """
        Called from `stpipe.Step.run` after `stpipe.Step.process` has
        completed. This method will call `stpipe.Step.finalize_result`
        for each model in the library.

        Parameters
        ----------
        step : Step
            The step calling this method.

        reference_files_used : list
            List of reference files used during the execution of this
            step.
        """
        with self:
            for i, model in enumerate(self):
                step.finalize_result(model, reference_files_used)
                self.shelve(model, i)

    def __enter__(self):
        """
        "open" the library. This is required for any "borrow" and "shelve"
        calls. While the library is "open" it will track which models are
        borrowed and raise a `BorrowError` if an attempt is made to close
        the library before all borrowed models are returned.
        """
        self._open = True
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        "close" the library.

        Raises
        ------
        `BorrowError`
            If the library "closes" and one or more models have not been
            "shelved" (returned) this exception will be raised.
        """
        self._open = False
        if exc_value:
            # if there is already an exception, don't worry about checking the ledger
            # instead allowing the calling code to raise the original error to provide
            # a more useful feedback without any chained ledger exception about
            # un-returned models
            return
        if self._ledger:
            raise BorrowError(
                f"ModelLibrary has {len(self._ledger)} un-returned models"
            )

    def map_function(self, function, modify=True):
        """
        Call a function once for each model in the library.

        Similar to the built-in ``map`` function this applies the
        function "function" to each model in the library by iterating
        through the models (yielding on each function call).

        Parameters
        ----------
        function : callable
            A function that accepts 2 arguments, ``model`` and ``index``

        modify : bool, optional, default=True
            For an "on_disk" library, temporary files will only be written
            when modify is True. For a library that is not "on_disk"
            this option has no effect and any modifications made to the model
            while it was borrowed will persist.

        Returns
        -------
        result_iter : generator
            A generator that will return function results. This can be
            converted to a list ``list(result_iter)`` to get the results
            of all the function calls (ordered the same as the models in
            the library).
        """
        with self:
            for index, model in enumerate(self):
                try:
                    yield function(model, index)
                finally:
                    # this is in a finally to allow cleanup if the generator is
                    # deleted after it finishes (when it's not fully consumed)
                    self.shelve(model, index, modify)
                # remove the local reference to model here to allow it
                # to be garbage collected before the next model is generated
                del model

    def _model_to_filename(self, model):
        """
        Determine the "filename" for a model. This will be used
        when writing temporary files (if the library is "on_disk")
        and for the ``save`` method (see the method for a note about
        how this is not used for the pipeline code).

        By default this method will return ``model.meta.filename``
        or "model.asdf" (if ``model.meta.filename`` is None).

        Parameters
        ----------
        model : DataModel

        Returns
        -------
        filename : str
        """
        model_filename = model.meta.filename
        if model_filename is None:
            model_filename = "model.asdf"
        return model_filename

    def _to_group_id(self, model_or_filename, index):
        if isinstance(model_or_filename, AbstractDataModel):
            getter = self._model_to_group_id
        else:
            getter = self._filename_to_group_id

        try:
            return getter(model_or_filename)
        except NoGroupID:
            return f"exposure{index + 1:04d}"

    def _model_to_exptype(self, model):
        """
        Compute "exptype" from a model using the DataModel interface.

        This will be called for every model in the library:
            - when the library is created from a list of models
            - when _save is called
        In both cases the models are all in memory and this method
        can use the in memory DataModel to determine the "exptype"
        (likely ``model.meta.exptype``).

        Parameters
        ----------
        model : DataModel

        Returns
        -------
        exptype : str
            Exposure type (for example "SCIENCE").
        """
        return getattr(model.meta, "exptype", "SCIENCE")

    @property
    @abc.abstractmethod
    def crds_observatory(self):
        """
        Return the name of the observatory as a string

        Returns
        -------
        observatory : str
        """

    @abc.abstractmethod
    def _datamodels_open(self, filename, **kwargs):
        """
        Open a model from a filename

        Parameters
        ----------
        filename : str or Path
            Filename from which to load a model.

        **kwargs : dict, optional
            Arguments to be passed to any lower-level function that
            opens a model.
        """

    @classmethod
    @abc.abstractmethod
    def _load_asn(cls, filename):
        """
        Load an association from a filename

        Parameters
        ----------
        filename : str or Path
            Filename from which to load an association
        """

    @abc.abstractmethod
    def _filename_to_group_id(self, filename):
        """
        Compute a "group_id" for a DataModel in filename.

        This will be called for every member in the association
        when the library is created if the member entry does not
        contain a "group_id" (this is the most efficient mode).
        Ideally this method should avoid opening the filename as
        a DataModel and instead take "short-cuts" to read only
        the "group_id" (from FITS or ASDF headers).

        If no "group_id" can be determined `NoGroupID` should be
        raised (to allow the library to assign a unique "group_id").

        Parameters
        ----------
        filename : str or Path
            Filename containing a DataModel

        Returns
        -------
        group_id : str
            "group_id" (used for ``group_names`` and ``group_indices``)
            for the model in the provided filename.
        """

    @abc.abstractmethod
    def _model_to_group_id(self, model):
        """
        Compute a "group_id" from a model using the DataModel interface.

        This will be called for every model in the library:
            - when the library is created from a list of models
            - when _save is called
        In both cases the models are all in memory and this method
        can use the in memory DataModel to determine the "group_id"
        (likely ``model.meta.group_id``).

        If no "group_id" can be determined `NoGroupID` should be
        raised (to allow the library to assign a unique "group_id").

        Parameters
        ----------
        model : DataModel

        Returns
        -------
        group_id : str
            "group_id" (used for ``group_names`` and ``group_indices``)
            for the model in the provided filename.
        """


import setuptools
from pathlib import Path
spike_home = Path(__file__).parent
pypi_descrip = (spike_home / "README.md").read_text()

setuptools.setup(
	name = "spike-psf",
	version = "1.0.7",
	author = "Ava Polzin",
	author_email = "apolzin@uchicago.edu",
	description = "Drizzle/resample HST, JWST, and Roman PSFs for improved analyses.",
	packages = ["spike", "spike/psf", "spike/psfgen", "spike/tools", 
	"spike/jwstcal", "spike/romancal", "spike/stcal", "spike/stpipe"],
	url = "https://github.com/avapolzin/spike",
	license = 'MIT',
	classifiers = [
		"Development Status :: 5 - Production/Stable",
		"Intended Audience :: Science/Research",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
		"Programming Language :: Python",
		"Topic :: Scientific/Engineering :: Astronomy",
		"Topic :: Scientific/Engineering :: Physics"],
	python_requires = ">=3.10",
	install_requires = ["asdf", "astropy", "crds", "drizzle", 
	# "drizzlepac @ git+https://github.com/spacetelescope/drizzlepac.git", 
	"gwcs", 
	"jsonschema", "matplotlib", "numpy", "photutils", "psutil", "pyyaml", "roman-datamodels", 
	"scipy", "spherical-geometry", "stdatamodels==2.2.0", "tweakwcs", 
	"stpsf"],
	package_data={'spike': ['configs/*/*']},
	long_description=pypi_descrip,
    long_description_content_type='text/markdown'
)



"""
This module is designed to make romancal tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


romancal copyright notice:

Copyright (C) 2010 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/romancal/blob/main/romancal/associations/lib/process_list.py
"""

from collections import deque
from enum import Enum
from functools import reduce

__all__ = [
    "ListCategory",
    "ProcessList",
    "ProcessItem",
    "ProcessQueue",
    "ProcessQueueSorted",
]


class ListCategory(Enum):
    """Categories in the list"""

    RULES = 0
    BOTH = 1
    EXISTING = 2
    NONSCIENCE = 3


class ProcessItem:
    """Items to be processed

    Parameters
    ----------
    obj : object
        The object to make a `ProcessItem`.
        Objects must be equatable.
    """

    def __init__(self, obj):
        self.obj = obj

    @classmethod
    def to_process_items(cls, iterable):
        """Iterable to convert a list to ProcessItem's

        Parameters
        ----------
        iterable : iterable
            A source of objects to convert

        Returns
        -------
        An iterable where the object has been
        converted to a `ProcessItem`
        """
        for obj in iterable:
            yield cls(obj)

    def __hash__(self):
        try:
            hash_value = self.obj.__hash__()
        except (AttributeError, TypeError):
            hash_value = hash(repr(self))
        return hash_value

    def __eq__(self, other):
        try:
            equality = self.obj == other.obj
        except AttributeError:
            equality = self.__hash__() == other.__hash__()
        return equality


class ProcessList:
    """A Process list

    Parameters
    ----------
    items : [item[, ...]]
        The list of items to process

    rules : [Association[, ...]]
        List of rules to process the items against.

    work_over : int
        What the reprocessing should work on:
        - `ProcessList.EXISTING`:   Only existing associations
        - `ProcessList.RULES`:      Only on the rules to create new associations
        - `ProcessList.BOTH`:       Compare to both existing and rules
        - `ProcessList.NONSCIENCE`: Only on non-science items

    only_on_match : bool
        Only use this object if the overall condition
        is True.

    trigger_constraints : [Constraint[,...]]
        The constraints that created the ProcessList

    trigger_rules : [Association[,...]]
        The association rules that created the ProcessList
    """

    _str_attrs = (
        "rules",
        "work_over",
        "only_on_match",
        "trigger_constraints",
        "trigger_rules",
    )

    def __init__(
        self,
        items=None,
        rules=None,
        work_over=ListCategory.BOTH,
        only_on_match=False,
        trigger_constraints=None,
        trigger_rules=None,
    ):
        self.items = items
        self.rules = rules
        self.work_over = work_over
        self.only_on_match = only_on_match
        self.trigger_constraints = (
            set(trigger_constraints) if trigger_constraints else set()
        )
        self.trigger_rules = set(trigger_rules) if trigger_rules else set()

    @property
    def hash(self):
        """Create a unique hash"""
        return (tuple(self.rules), self.work_over, self.only_on_match)

    def update(self, process_list, full=False):
        """Update with information from ProcessList

        Attributes from `process_list` are added to self's attributes. If `not
        full`, the attributes `rules`, 'work_over`, and `only_on_match` are not
        taken.

        Note that if `full`, destructive action will occur with respect to
        `work_over` and `only_on_match`.

        Parameters
        ----------
        process_list : ProcessList
            The source process list to absorb.

        full : bool
            Include the hash attributes `rules`, `work_over`, and `only_on_match`.
        """
        self.items += process_list.items
        self.trigger_constraints.update(process_list.trigger_constraints)
        self.trigger_rules.update(process_list.trigger_rules)
        if full:
            self.rules += process_list.rules
            self.work_over = process_list.work_over
            self.only_on_match = process_list.only_on_match

    def __str__(self):
        result = "{}(n_items: {}, {})".format(
            self.__class__.__name__,
            len(self.items),
            {str_attr: getattr(self, str_attr) for str_attr in self._str_attrs},
        )
        return result


class ProcessQueue(deque):
    """Make a deque iterable and mutable"""

    def __iter__(self):
        while True:
            try:
                yield self.popleft()
            except (IndexError, StopIteration):
                break


class ProcessListQueue:
    """First-In-First-Out queue of ProcessLists

    Parameters
    ----------
    init : [ProcessList[,...]] or None
        List of ProcessLists to put on the queue.
    """

    def __init__(self, init=None):
        self._queue = {}
        if init is not None:
            self.extend(init)

    def append(self, process_list):
        """Add ProcessList to queue"""
        plhash = process_list.hash
        if plhash not in self._queue:
            self._queue[plhash] = process_list
        else:
            self._queue[plhash].update(process_list)

    def extend(self, iterable):
        """Add objects if not already in the queue"""
        for process_list in iterable:
            self.append(process_list)

    def items(self):
        """Return list generator of all items"""
        for plhash in self._queue:
            yield from self._queue[plhash].items

    def popleft(self):
        """Pop the first-in object"""
        plhash = next(iter(self._queue))
        process_list = self._queue[plhash]
        del self._queue[plhash]
        return process_list

    def __len__(self):
        return len(self._queue)

    def __iter__(self):
        while True:
            try:
                yield self.popleft()
            except (IndexError, StopIteration):
                break

    def __str__(self):
        result = (
            f"{self.__class__.__name__}: rulesets {len(self)} items"
            f" {len(list(self.items()))}"
        )
        return result


class ProcessQueueSorted:
    """Sort ProcessItem based on work_over

    `ProcessList`s are handled in order of `RULES`, `BOTH`,
    `EXISTING`, and `NONSCIENCE`.

    Parameters
    ----------
    init : [ProcessList[,...]]
        List of `ProcessList` to start the queue with.
    """

    def __init__(self, init=None):
        self.queues = {
            list_category: ProcessListQueue() for list_category in ListCategory
        }

        if init is not None:
            self.extend(init)

    def extend(self, process_lists):
        """Add the list of process items to their appropriate queues"""
        for process_list in process_lists:
            self.queues[process_list.work_over].append(process_list)

    def __iter__(self):
        """Return the queues in order"""
        while len(self) > 0:
            for category in ListCategory:
                for process_list in self.queues[category]:
                    yield process_list
                    break
                else:
                    continue
                break

    def __len__(self):
        return reduce(lambda x, y: x + len(y), self.queues.values(), 0)

    def __str__(self):
        result = f"{self.__class__.__name__}:"
        for queue in self.queues:
            result += f"\n\tQueue {queue}: {self.queues[queue]}"
        return result


def workover_filter(process_list, work_over):
    """Determine and modify workover of input process list

    Parameters
    ---------
    process_list : ProcessList
        The process list under consideration

    work_over : ListCategory
        The ListCategory to compare against.

    Returns
    -------
    process_list : ProcessList or None
        The input process_list with work_over modified.
        None if the process list should not be continued.
    """
    result = process_list
    if process_list.work_over in [ListCategory.RULES, ListCategory.BOTH]:
        if work_over in [ListCategory.RULES, ListCategory.BOTH]:
            result.work_over = ListCategory.BOTH
        else:
            result.work_over = work_over
    else:
        if work_over not in [ListCategory.RULES, ListCategory.BOTH]:
            result = None
    return result


"""
This module is designed to make romancal tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


romancal copyright notice:

Copyright (C) 2010 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/romancal/blob/main/romancal/datamodels/library.py
"""

import asdf
from roman_datamodels import open as datamodels_open
from spike.stpipe.library import AbstractModelLibrary, NoGroupID

# from romancal.associations import AssociationNotValidError, load_asn
from spike.romancal.associations_exceptions import AssociationNotValidError
from spike.romancal.associations_load_asn import load_asn

__all__ = ["ModelLibrary"]


class ModelLibrary(AbstractModelLibrary):
    @property
    def crds_observatory(self):
        return "roman"

    def _model_to_filename(self, model):
        model_filename = model.meta.filename
        if model_filename is None:
            model_filename = "model.asdf"
        return model_filename

    def _datamodels_open(self, filename, **kwargs):
        return datamodels_open(filename, **kwargs)

    @classmethod
    def _load_asn(cls, asn_path):
        try:
            with open(asn_path) as asn_file:
                asn_data = load_asn(asn_file)
        except AssociationNotValidError as e:
            raise OSError("Cannot read ASN file.") from e
        return asn_data

    def _filename_to_group_id(self, filename):
        """
        Compute a "group_id" without loading the file as a DataModel

        This function will return the meta.group_id stored in the ASDF
        extension (if it exists) or a group_id calculated from the
        ASDF headers.
        """
        meta = asdf.util.load_yaml(filename)["roman"]["meta"]
        if group_id := meta.get("group_id"):
            return group_id
        if "observation" in meta:
            return _mapping_to_group_id(meta["observation"])
        raise NoGroupID(f"{filename} missing group_id")

    def _model_to_group_id(self, model):
        """
        Compute a "group_id" from a model using the DataModel interface
        """
        if (group_id := getattr(model.meta, "group_id", None)) is not None:
            return group_id
        if hasattr(model.meta, "observation"):
            return _mapping_to_group_id(model.meta.observation)
        raise NoGroupID(f"{model} missing group_id")

    def _assign_member_to_model(self, model, member):
        # roman_datamodels doesn't allow assignment of meta.group_id
        # (since it's not in the schema). To work around this use
        # __setitem__ calls here instead of setattr
        for attr in ("group_id", "tweakreg_catalog", "exptype"):
            if attr in member:
                model.meta[attr] = member[attr]
        if not hasattr(model.meta, "asn"):
            model.meta["asn"] = {}

        if "table_name" in self.asn:
            model.meta.asn["table_name"] = self.asn["table_name"]
        if "asn_pool" in self.asn:
            model.meta.asn["pool_name"] = self.asn["asn_pool"]


def _mapping_to_group_id(mapping):
    """
    Combine a number of file metadata values into a ``group_id`` string
    """
    return "{observation_id}".format_map(mapping)


"""
This module is designed to make jwst tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


jwst copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/jwst/blob/main/jwst/tweakreg/tweakreg_catalog.py
"""

import logging
import inspect
import warnings

from astropy.table import Table
from astropy.utils.exceptions import AstropyUserWarning
import numpy as np
from photutils.detection import DAOStarFinder, IRAFStarFinder
from photutils.segmentation import SourceFinder, SourceCatalog

from stdatamodels.jwst.datamodels import dqflags, ImageModel

from spike.jwstcal.source_catalog_detection import JWSTBackground

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def _SourceFinderWrapper(data, threshold, mask=None, **kwargs):
    """
    Wrapper function for photutils.source_finder.SourceFinder to make input
    and output consistent with DAOStarFinder and IRAFStarFinder.
    see `photutils segmentation tutorial <https://photutils.readthedocs.io/en/stable/segmentation.html>`_.

    Parameters
    ----------
    data : array_like
        The 2D array of the image.
    threshold : float
        The absolute image value above which to select sources.
    mask : array_like (bool), optional
        The image mask
    **kwargs : Additional keyword arguments are passed to `photutils.segmentation.SourceFinder`
        and/or `photutils.segmentation.SourceCatalog`.

    Returns
    -------
    sources : `~astropy.table.QTable`
        A table containing the found sources.
    """

    default_kwargs = {'npixels': 10,
                      'progress_bar': False,
                      }
    kwargs = {**default_kwargs, **kwargs}

    # handle passing kwargs into SourceFinder and SourceCatalog
    # note that this suppresses TypeError: unexpected keyword arguments
    # so user must be careful to know which kwargs are passed in here
    finder_args = list(inspect.signature(SourceFinder).parameters)
    catalog_args = list(inspect.signature(SourceCatalog).parameters)
    finder_dict = {k: kwargs.pop(k) for k in dict(kwargs) if k in finder_args}
    catalog_dict = {k: kwargs.pop(k) for k in dict(kwargs) if k in catalog_args}
    if ('kron_params' in catalog_dict.keys()) and (catalog_dict['kron_params'] is None):
        catalog_dict['kron_params'] = (2.5, 1.4, 0.0)  # necessary because cannot specify default in Step spec string

    finder = SourceFinder(**finder_dict)
    segment_map = finder(data, threshold, mask=mask)
    sources = SourceCatalog(data, segment_map, mask=mask, **catalog_dict).to_table()
    sources.rename_column('label', 'id')
    sources.rename_column('segment_flux', 'flux')

    return sources


def _IRAFStarFinderWrapper(data, threshold, mask=None, **kwargs):
    """
    Wrapper function for `photutils.detection.IRAFStarFinder` to make inputs
    and outputs consistent across the three allowed detection methods

    Parameters
    ----------
    data : array_like
        The 2D array of the image.
    threshold : float
        The absolute image value above which to select sources.
    mask : array_like (bool), optional
        The image mask
    **kwargs : Additional keyword arguments are passed to
        `photutils.detection.IRAFStarFinder`.

    Returns
    -------
    sources : `~astropy.table.QTable`
        A table containing the found sources.
    """

    # defaults are not necessary to repeat here when running full pipeline step
    # but direct call to make_tweakreg_catalog will fail without 'fwhm' specified
    default_kwargs = {'fwhm': 2.5,}
    kwargs = {**default_kwargs, **kwargs}

    # note that this suppresses TypeError: unexpected keyword arguments
    # so user must be careful to know which kwargs are passed in here
    finder_args = list(inspect.signature(IRAFStarFinder).parameters)
    finder_dict = {k: kwargs.pop(k) for k in dict(kwargs) if k in finder_args}
    fwhm = finder_dict.pop('fwhm')

    starfind = IRAFStarFinder(threshold, fwhm, **finder_dict)
    sources = starfind(data, mask=mask)

    return sources


def _DaoStarFinderWrapper(data, threshold, mask=None, **kwargs):
    """
    Wrapper function for `photutils.detection.DAOStarFinder` to make inputs
    and outputs consistent across the three allowed detection methods

    Parameters
    ----------
    data : array_like
        The 2D array of the image.
    threshold : float
        The absolute image value above which to select sources.
    mask : array_like (bool), optional
        The image mask
    **kwargs : Additional keyword arguments are passed to
        a`photutils.detection.DAOStarFinder`.

    Returns
    -------
    sources : `~astropy.table.QTable`
        A table containing the found sources.
    """

    # defaults are not necessary to repeat here when running full pipeline step
    # but direct call to make_tweakreg_catalog will fail without 'fwhm' specified
    default_kwargs = {'fwhm': 2.5,}
    kwargs = {**default_kwargs, **kwargs}

    # for consistency with IRAFStarFinder, allow minsep_fwhm to be passed in
    # and convert to pixels in the same way that IRAFStarFinder does
    # see IRAFStarFinder readthedocs page and also
    # https://github.com/astropy/photutils/issues/1561
    if "minsep_fwhm" in kwargs:
        min_sep_pix = max(2, int(kwargs["minsep_fwhm"] * kwargs["fwhm"] + 0.5))
        kwargs["min_separation"] = min_sep_pix

    # note that this suppresses TypeError: unexpected keyword arguments
    # so user must be careful to know which kwargs are passed in here
    finder_args = list(inspect.signature(DAOStarFinder).parameters)
    finder_dict = {k: kwargs.pop(k) for k in dict(kwargs) if k in finder_args}
    fwhm = finder_dict.pop('fwhm')
    starfind = DAOStarFinder(threshold, fwhm, **finder_dict)
    sources = starfind(data, mask=mask)

    return sources


def make_tweakreg_catalog(model, snr_threshold, bkg_boxsize=400, starfinder='iraf', starfinder_kwargs={}):
    """
    Create a catalog of point-line sources to be used for image
    alignment in tweakreg.

    Parameters
    ----------
    model : `ImageModel`
        The input `ImageModel` of a single image.  The input image is
        assumed to be background subtracted.

    snr_threshold : float
        The signal-to-noise ratio per pixel above the ``background`` for
        which to consider a pixel as possibly being part of a source.

    bkg_boxsize : float, optional
        The background mesh box size in pixels.

    starfinder : str, optional
        The `photutils` star finder to use.  Options are 'dao', 'iraf', or 'segmentation'.

            - 'dao': `photutils.detection.DAOStarFinder`
            - 'iraf': `photutils.detection.IRAFStarFinder`
            - 'segmentation': `photutils.segmentation.SourceFinder` and `photutils.segmentation.SourceCatalog`

    starfinder_kwargs : dict, optional
        additional keyword arguments to be passed to the star finder.
        for 'segmentation', these can be kwargs to `photutils.segmentation.SourceFinder`
        and/or `photutils.segmentation.SourceCatalog`.
        for 'dao' or 'iraf', these are kwargs to `photutils.detection.DAOStarFinder`
        or `photutils.detection.IRAFStarFinder`, respectively.
        Defaults are as stated in the docstrings of those functions unless noted here:

            - 'dao': fwhm=2.5
            - 'iraf': fwhm=2.5
            - 'segmentation': npixels=10, progress_bar=False

    Returns
    -------
    catalog : `~astropy.Table`
        An astropy Table containing the source catalog.
    """
    if not isinstance(model, ImageModel):
        raise TypeError('The input model must be an ImageModel.')

    if starfinder.lower() in ['dao', 'daostarfinder']:
        StarFinder = _DaoStarFinderWrapper
    elif starfinder.lower() in ['iraf', 'irafstarfinder']:
        StarFinder = _IRAFStarFinderWrapper
    elif starfinder.lower() in ['segmentation', 'sourcefinder']:
        StarFinder = _SourceFinderWrapper
    else:
        raise ValueError(f"Unknown starfinder type: {starfinder}")

    # Mask the non-imaging area, e.g. reference pixels and MIRI non-science area
    coverage_mask = ((dqflags.pixel['NON_SCIENCE'] +
                      dqflags.pixel['DO_NOT_USE']) &
                     model.dq).astype(bool)

    columns = ['id', 'xcentroid', 'ycentroid', 'flux']
    try:
        bkg = JWSTBackground(model.data, box_size=bkg_boxsize,
                             coverage_mask=coverage_mask)
        with warnings.catch_warnings():
            # suppress warning about NaNs being automatically masked - this is desired
            warnings.simplefilter('ignore', AstropyUserWarning) 
            threshold_img = bkg.background + (snr_threshold * bkg.background_rms)
    except ValueError as e:
        log.warning(f"Error determining sky background: {e.args[0]}")
        # return an empty catalog
        catalog = Table(names=columns, dtype=(int, float, float, float))
        return catalog

    threshold = np.median(threshold_img)  # DAOStarFinder requires float
    sources = StarFinder(model.data, threshold, mask=coverage_mask, **starfinder_kwargs)

    if sources:
        catalog = sources[columns]
    else:
        # return an empty table
        catalog = Table(names=columns, dtype=(int, float, float, float))

    return catalog